"""Handler utilities for getting non-monkey patched standard library stuff.

Allows one to get an unpatched thread module, with a thread
decorator that uses the unpatching OS thread.

"""
from __future__ import absolute_import

try:
    from gevent._threading import start_new_thread
except ImportError:
    from thread import start_new_thread


def thread(func):
    """Thread decorator

    Takes a function and spawns it as a daemon thread using the
    real OS thread regardless of monkey patching.

    """
    start_new_thread(func, ())



"""Zookeeper Partitioner Implementation

:class:`SetPartitioner` implements a partitioning scheme using
Zookeeper for dividing up resources amongst members of a party.

This is useful when there is a set of resources that should only be
accessed by a single process at a time that multiple processes
across a cluster might want to divide up.

Example Use-Case
----------------

- Multiple workers across a cluster need to divide up a list of queues
  so that no two workers own the same queue.

"""
import logging
import os
import socket
import time
import threading
from functools import partial

from kazoo.client import KazooState


log = logging.getLogger(__name__)


class PartitionState(object):
    """High level partition state values

    .. attribute:: ALLOCATING

        The set needs to be partitioned, and may require an existing
        partition set to be released before acquiring a new partition
        of the set.

    .. attribute:: ACQUIRED

        The set has been partitioned and acquired.

    .. attribute:: RELEASE

        The set needs to be repartitioned, and the current partitions
        must be released before a new allocation can be made.

    .. attribute:: FAILURE

        The set partition has failed. This occurs when the maximum
        time to partition the set is exceeded or the Zookeeper session is
        lost. The partitioner is unusable after this state and must
        be recreated.

    """
    ALLOCATING = "ALLOCATING"
    ACQUIRED = "ACQUIRED"
    RELEASE = "RELEASE"
    FAILURE = "FAILURE"


class SetPartitioner(object):
    """Partitions a set amongst members of a party

    This class will partition a set amongst members of a party such
    that each member will be given zero or more items of the set and
    each set item will be given to a single member. When new members
    enter or leave the party, the set will be re-partitioned amongst the
    members.

    When the :class:`SetPartitioner` enters the
    :attr:`~PartitionState.FAILURE` state, it is unrecoverable
    and a new :class:`SetPartitioner` should be created.


    Simple Example:

    .. code-block:: python

        from kazoo.client import KazooClient
        client = KazooClient()

        qp = client.SetPartitioner(
            path='/work_queues', set=('queue-1', 'queue-2', 'queue-3'))

        def use_setlist(partitions):
            # do something with a partition passed in

        # Run the use_setlist function when partitions are acquired
        # repeatedly
        qp.run(func=use_setlist)

    Sometimes, more control is needed over handling the state
    transitions. Or the program may need to do other things during
    specific states, in which case a more verbose example that allows
    for finer grained control is needed:

    .. code-block:: python

        from kazoo.client import KazooClient
        client = KazooClient()

        qp = client.SetPartitioner(
            path='/work_queues', set=('queue-1', 'queue-2', 'queue-3'))

        while 1:
            if qp.failed:
                raise Exception("Lost or unable to acquire partition")
            elif qp.release:
                qp.release_set()
            elif qp.acquired:
                for partition in qp:
                    # Do something with each partition
            elif qp.allocating:
                qp.wait_for_acquire()

    **State Transitions**

    When created, the :class:`SetPartitioner` enters the
    :attr:`PartitionState.ALLOCATING` state.

    :attr:`~PartitionState.ALLOCATING` ->
    :attr:`~PartitionState.ACQUIRED`

        Set was partitioned successfully, the partition list assigned
        is accessible via list/iter methods or calling list() on the
        :class:`SetPartitioner` instance.

    :attr:`~PartitionState.ALLOCATING` ->
    :attr:`~PartitionState.FAILURE`

        Allocating the set failed either due to a Zookeeper session
        expiration, or failure to acquire the items of the set within
        the timeout period.

    :attr:`~PartitionState.ACQUIRED` ->
    :attr:`~PartitionState.RELEASE`

        The members of the party have changed, and the set needs to be
        repartitioned. :meth:`SetPartitioner.release` should be called
        as soon as possible.

    :attr:`~PartitionState.ACQUIRED` ->
    :attr:`~PartitionState.FAILURE`

        The current partition was lost due to a Zookeeper session
        expiration.

    :attr:`~PartitionState.RELEASE` ->
    :attr:`~PartitionState.ALLOCATING`

        The current partition was released and is being re-allocated.

    """
    def __init__(self, client, path, set, partition_func=None,
                 identifier=None, time_boundary=30):
        """Create a :class:`~SetPartitioner` instance

        :param client: A :class:`~kazoo.client.KazooClient` instance
        :param path: The partition path to use
        :param set: The set of items to partition
        :param partition_func: A function to use to decide how to
                               partition the set
        :param identifier: An identifier to use for this member of the
                           party when participating. Defaults to the
                           hostname + process id.
        :param time_boundary: How long the party members must be stable
                              before allocation can complete

        """
        self.state = PartitionState.ALLOCATING

        self._client = client
        self._path = path
        self._set = set
        self._partition_set = []
        self._partition_func = partition_func or self._partitioner
        self._identifier = identifier or '%s-%s' % (
            socket.getfqdn(), os.getpid())
        self._locks = []
        self._lock_path = '/'.join([path, 'locks'])
        self._party_path = '/'.join([path, 'party'])
        self._time_boundary = time_boundary

        self._acquire_event = client._handler.event_object()

        # Create basic path nodes
        client.ensure_path(path)
        client.ensure_path(self._lock_path)
        client.ensure_path(self._party_path)

        # Join the party
        self._party = client.ShallowParty(self._party_path,
                                          identifier=self._identifier)
        self._party.join()

        self._state_change = threading.Condition()
        client.add_listener(self._establish_sessionwatch)

        # Now watch the party and set the callback on the async result
        # so we know when we're ready
        self._children_updated = False
        self._child_watching(self._allocate_transition, async=True)

    def __iter__(self):
        """Return the partitions in this partition set"""
        for partition in self._partition_set:
            yield partition

    @property
    def failed(self):
        """Corresponds to the :attr:`PartitionState.FAILURE` state"""
        return self.state == PartitionState.FAILURE

    @property
    def release(self):
        """Corresponds to the :attr:`PartitionState.RELEASE` state"""
        return self.state == PartitionState.RELEASE

    @property
    def allocating(self):
        """Corresponds to the :attr:`PartitionState.ALLOCATING`
        state"""
        return self.state == PartitionState.ALLOCATING

    @property
    def acquired(self):
        """Corresponds to the :attr:`PartitionState.ACQUIRED` state"""
        return self.state == PartitionState.ACQUIRED

    def wait_for_acquire(self, timeout=30):
        """Wait for the set to be partitioned and acquired

        :param timeout: How long to wait before returning

        """
        self._acquire_event.wait(timeout)

    def release_set(self):
        """Call to release the set

        This method begins the step of allocating once the set has
        been released.

        """
        self._release_locks()
        if self._locks:
            # This shouldn't happen, it means we couldn't release our
            # locks, abort
            self._fail_out()
            return
        else:
            with self._state_change:
                if self.failed:
                    return
                self.state = PartitionState.ALLOCATING
        self._child_watching(self._allocate_transition, async=True)

    def finish(self):
        """Call to release the set and leave the party"""
        self._release_locks()
        self._fail_out()

    def _fail_out(self):
        with self._state_change:
            self.state = PartitionState.FAILURE
        if self._party.participating:
            try:
                self._party.leave()
            except Exception:
                pass

    def _allocate_transition(self, result):
        """Called when in allocating mode, and the children settled"""
        # Did we get an exception waiting for children to settle?
        if result.exception:
            self._fail_out()
            return

        children, async_result = result.get()
        self._children_updated = False

        # Add a callback when children change on the async_result
        def updated(result):
            with self._state_change:
                if self.acquired:
                    self.state = PartitionState.RELEASE
            self._children_updated = True

        async_result.rawlink(updated)

        # Split up the set
        self._partition_set = self._partition_func(
            self._identifier, list(self._party), self._set)

        # Proceed to acquire locks for the working set as needed
        for member in self._partition_set:
            if self._children_updated or self.failed:
                # Still haven't settled down, release locks acquired
                # so far and go back
                return self._abort_lock_acquisition()

            lock = self._client.Lock(self._lock_path + '/' +
                                     str(member))
            try:
                lock.acquire()
            except Exception:
                return self.finish()
            self._locks.append(lock)

        # All locks acquired! Time for state transition, make sure
        # we didn't inadvertently get lost thus far
        with self._state_change:
            if self.failed:
                return self.finish()
            self.state = PartitionState.ACQUIRED
            self._acquire_event.set()

    def _release_locks(self):
        """Attempt to completely remove all the locks"""
        self._acquire_event.clear()
        for lock in self._locks[:]:
            try:
                lock.release()
            except Exception:
                # We proceed to remove as many as possible, and leave
                # the ones we couldn't remove
                pass
            else:
                self._locks.remove(lock)

    def _abort_lock_acquisition(self):
        """Called during lock acquisition if a party change occurs"""
        self._partition_set = None
        self._release_locks()
        if self._locks:
            # This shouldn't happen, it means we couldn't release our
            # locks, abort
            self._fail_out()
            return
        return self._child_watching(self._allocate_transition)

    def _child_watching(self, func=None, async=False):
        """Called when children are being watched to stabilize

        This actually returns immediately, child watcher spins up a
        new thread/greenlet and waits for it to stabilize before
        any callbacks might run.

        """
        watcher = ChildrenWatcher(self._client, self._party_path,
                                  self._time_boundary)
        asy = watcher.start()
        if func:
            # We spin up the function in a separate thread/greenlet
            # to ensure that the rawlink's it might use won't be
            # blocked
            if async:
                func = partial(self._client._handler.spawn, func)
            asy.rawlink(func)
        return asy

    def _establish_sessionwatch(self, state):
        """Register ourself to listen for session events, we shut down
        if we become lost"""
        if state == KazooState.LOST:
            self._client.remove_listener(self._establish_sessionwatch)
            self._fail_out()

    def _partitioner(self, identifier, members, partitions):
        # Ensure consistent order of partitions/members
        all_partitions = sorted(partitions)
        workers = sorted(members)

        i = workers.index(identifier)
        # Now return the partition list starting at our location and
        # skipping the other workers
        return all_partitions[i::len(workers)]


class ChildrenWatcher(object):
    """Node Children Watcher that alerts when the children stabilize

    A separate watcher for the children of a node, that ignores
    changes within a boundary time and sets the result only when the
    boundary time has elapsed and there is no change of children for
    the boundary time's duration.

    Example::

        watcher = ChildrenWatcher(client, '/some/path')
        result = watcher.start()

        # Blocks until the children have not changed for time boundary
        # seconds, returns list of children and an async_result that
        # will be set once if the children change in the future
        children, async_result = result.get()

    """
    def __init__(self, client, path, time_boundary=30):
        self.client = client
        self.path = path
        self.time_boundary = time_boundary
        self.children = []
        self.last_change = None

    def start(self):
        """Begin the watching process asynchronously

        :returns: An :class:`~kazoo.interfaces.IAsyncResult` instance
                  that will be set when no change has occurred to the
                  children for time boundary seconds.

        """
        self.asy = asy = self.client._handler.async_result()
        self.client._handler.spawn(self._inner_start)
        return asy

    def _inner_start(self):
        if not self.last_change:
            self.last_change = time.time()

        try:
            async_result = None
            while time.time() - self.last_change < self.time_boundary:
                async_result = self.client._handler.async_result()
                self.children = self.client.retry(
                    self.client.get_children, self.path,
                    partial(self.children_watcher, async_result))

                # Now sleep for a time boundary's worth from the last change
                sleep_till = self.time_boundary - (time.time() -
                                                   self.last_change)
                time.sleep(sleep_till)

            self.asy.set((self.children, async_result))
        except Exception as exc:
            self.asy.set_exception(exc)

    def children_watcher(self, async, event):
        self.last_change = time.time()
        async.set(time.time())



"""Threading Handler

The :class:`SequentialThreadingHandler` is intended for regular Python
environments that use threads.

.. warning::

    Do not use :class:`SequentialThreadingHandler` with applications using
    asynchronous event loops (like gevent). Use the
    :class:`~kazoo.handlers.gevent.SequentialGeventHandler` instead.

"""
from __future__ import absolute_import

import Queue
import time
import threading

from zope.interface import implementer

from kazoo.handlers.util import thread
from kazoo.interfaces import IAsyncResult
from kazoo.interfaces import IHandler

# sentinal object
_NONE = object()


class TimeoutError(Exception):
    pass


@implementer(IAsyncResult)
class AsyncResult(object):
    """A one-time event that stores a value or an exception"""
    def __init__(self, handler):
        self._handler = handler
        self.value = None
        self._exception = _NONE
        self._condition = threading.Condition()
        self._callbacks = []

    def ready(self):
        """Return true if and only if it holds a value or an exception"""
        return self._exception is not _NONE

    def successful(self):
        """Return true if and only if it is ready and holds a value"""
        return self._exception is None

    @property
    def exception(self):
        if self._exception is not _NONE:
            return self._exception

    def set(self, value=None):
        """Store the value. Wake up the waiters."""
        with self._condition:
            self.value = value
            self._exception = None

            for callback in self._callbacks:
                self._handler.completion_queue.put(
                    lambda: callback(self)
                )
            self._condition.notify_all()

    def set_exception(self, exception):
        """Store the exception. Wake up the waiters."""
        with self._condition:
            self._exception = exception

            for callback in self._callbacks:
                self._handler.completion_queue.put(
                    lambda: callback(self)
                )
            self._condition.notify_all()

    def get(self, block=True, timeout=None):
        """Return the stored value or raise the exception.

        If there is no value raises TimeoutError.

        """
        with self._condition:
            if self._exception is not _NONE:
                if self._exception is None:
                    return self.value
                raise self._exception
            elif block:
                self._condition.wait(timeout)
                if self._exception is not _NONE:
                    if self._exception is None:
                        return self.value
                    raise self._exception

            # if we get to this point we timeout
            raise TimeoutError()

    def get_nowait(self):
        """Return the value or raise the exception without blocking.

        If nothing is available, raises TimeoutError

        """
        return self.get(block=False)

    def rawlink(self, callback):
        """Register a callback to call when a value or an exception is set"""
        with self._condition:
            # Are we already set? Dispatch it now
            if self.ready():
                self._handler.completion_queue.put(
                    lambda: callback(self)
                )
                return

            if callback not in self._callbacks:
                self._callbacks.append(callback)

    def unlink(self, callback):
        """Remove the callback set by :meth:`rawlink`"""
        with self._condition:
            if self.ready():
                # Already triggered, ignore
                return

            if callback in self._callbacks:
                self._callbacks.remove(callback)


@implementer(IHandler)
class SequentialThreadingHandler(object):
    """threading Handler for sequentially executing callbacks

    This handler executes callbacks in a sequential manner from the Zookeeper
    thread. A queue is created for each of the callback events, so that each
    type of event has its callback type run sequentially. These are split into
    three queues, therefore it's possible that a session event arriving after a
    watch event may have its callback executed at the same time or slightly
    before the watch event callback.

    Each queue type has a thread worker that pulls the callback event off the
    queue and runs it in the order Zookeeper sent it.

    This split helps ensure that watch callbacks won't block session
    re-establishment should the connection be lost during a Zookeeper client
    call.

    Watch, session, and completion callbacks should avoid blocking behavior as
    the next callback of that type won't be run until it completes. If you need
    to block, spawn a new thread and return immediately so callbacks can
    proceed.

    .. note::

        Completion callbacks can block to wait on Zookeeper calls, but no
        other completion callbacks will execute until the callback returns.

    """
    name = "sequential_threading_handler"
    timeout_exception = TimeoutError
    sleep_func = time.sleep

    def __init__(self):
        """Create a :class:`SequentialThreadingHandler` instance"""
        self.callback_queue = Queue.Queue()
        self.session_queue = Queue.Queue()
        self.completion_queue = Queue.Queue()
        self._running = True

        # Spawn our worker threads, we have
        # - A callback worker for watch events to be called
        # - A session worker for session events to be called
        # - A completion worker for completion events to be called
        self._create_thread_worker(self.callback_queue)
        self._create_thread_worker(self.session_queue)
        self._create_thread_worker(self.completion_queue)

    def _create_thread_worker(self, queue):
        @thread
        def thread_worker():  # pragma: nocover
            while self._running:
                try:
                    func = queue.get(timeout=1)
                    func()
                except Queue.Empty:
                    continue

    def event_object(self):
        """Create an appropriate Event object"""
        return threading.Event()

    def lock_object(self):
        """Create a lock object"""
        return threading.Lock()

    def async_result(self):
        """Create a :class:`AsyncResult` instance"""
        return AsyncResult(self)

    def spawn(self, func, *args, **kwargs):
        t = threading.Thread(target=func, args=args, kwargs=kwargs)
        t.daemon = True
        t.start()

    def dispatch_callback(self, callback):
        """Dispatch to the callback object

        The callback is put on separate queues to run depending on the type
        as documented for the :class:`SequentialThreadingHandler`.

        """
        if callback.type == 'session':
            self.session_queue.put(lambda: callback.func(*callback.args))
        else:
            self.callback_queue.put(lambda: callback.func(*callback.args))



import unittest

from gevent.event import Event
from nose.tools import eq_

from kazoo.client import Callback
from kazoo.testing import KazooTestCase


class TestGeventHandler(unittest.TestCase):
    def _makeOne(self, *args):
        from kazoo.handlers.gevent import SequentialGeventHandler
        return SequentialGeventHandler(*args)

    def _getAsync(self, *args):
        from kazoo.handlers.gevent import AsyncResult
        return AsyncResult

    def test_completion_vs_session(self):
        h = self._makeOne()

        lst = []
        av = Event()
        bv = Event()
        cv = Event()
        dv = Event()

        def addToList():
            av.set()
            bv.wait()
            lst.append(True)
            dv.set()

        def anotherAdd():
            lst.append(True)
            cv.set()

        call1 = Callback('session', addToList, ())
        call2 = Callback('completion', anotherAdd, ())

        h.dispatch_callback(call1)
        av.wait()
        # Now we know the first is waiting, make sure
        # the second executes while the first has blocked
        # its thread
        h.dispatch_callback(call2)
        cv.wait()

        eq_(lst, [True])
        bv.set()
        dv.wait()
        eq_(lst, [True, True])

    def test_proper_threading(self):
        h = self._makeOne()
        assert isinstance(h.event_object(), Event)

    def test_matching_async(self):
        h = self._makeOne()
        async = self._getAsync()
        assert isinstance(h.async_result(), async)


class TestGeventClient(KazooTestCase):
    def _makeOne(self, *args):
        from kazoo.handlers.gevent import SequentialGeventHandler
        return SequentialGeventHandler(*args)

    def test_connect(self):
        client = self._get_client(handler=self._makeOne())
        client.connect()
        assert client.state == 'CONNECTED'
        client.stop()



"""Kazoo Zookeeper Client"""
import inspect
import logging
import os
from collections import namedtuple
from functools import partial
from os.path import split

import zookeeper

from kazoo.exceptions import BadArgumentsException
from kazoo.exceptions import ConfigurationError
from kazoo.exceptions import ZookeeperStoppedError
from kazoo.exceptions import NoNodeException
from kazoo.exceptions import err_to_exception
from kazoo.handlers.threading import SequentialThreadingHandler
from kazoo.recipe.lock import Lock
from kazoo.recipe.party import Party
from kazoo.recipe.party import ShallowParty
from kazoo.recipe.election import Election
from kazoo.retry import KazooRetry
from kazoo.handlers.util import thread

log = logging.getLogger(__name__)

ZK_OPEN_ACL_UNSAFE = {"perms": zookeeper.PERM_ALL, "scheme": "world",
                       "id": "anyone"}


## Zookeeper Logging Setup
# Setup Zookeeper logging thread
_logging_pipe = os.pipe()
zookeeper.set_log_stream(os.fdopen(_logging_pipe[1], 'w'))


@thread
def _loggingthread():
    """Zookeeper logging redirect

    Zookeeper by default logs directly out. This thread handles reading
    off the pipe that the above `set_log_stream` call designates so
    that the Zookeeper logging output can be turned into Python logging
    statements under the `Zookeeper` name.

    """
    r, w = _logging_pipe
    log = logging.getLogger('ZooKeeper').log
    f = os.fdopen(r)
    levels = dict(ZOO_INFO=logging.INFO,
                  ZOO_WARN=logging.WARNING,
                  ZOO_ERROR=logging.ERROR,
                  ZOO_DEBUG=logging.DEBUG,
                  )
    while 1:
        line = f.readline().strip()
        try:
            if '@' in line:
                level, message = line.split('@', 1)
                level = levels.get(level.split(':')[-1])

                if 'Exceeded deadline by' in line and level == logging.WARNING:
                    level = logging.DEBUG

            else:
                level = None

            if level is None:
                log(logging.INFO, line)
            else:
                log(level, message)
        except Exception as v:
            logging.getLogger('ZooKeeper').exception("Logging error: %s", v)


## Client State and Event objects

class KazooState(object):
    """High level connection state values

    States inspired by Netflix Curator.

    .. attribute:: SUSPENDED

        The connection has been lost but may be recovered. We should
        operate in a "safe mode" until then.

    .. attribute:: CONNECTED

        The connection is alive and well.

    .. attribute:: LOST

        The connection has been confirmed dead. Any ephemeral nodes
        will need to be recreated upon re-establishing a connection.

    """
    SUSPENDED = "SUSPENDED"
    CONNECTED = "CONNECTED"
    LOST = "LOST"


class KeeperState(object):
    """Zookeeper State

    Represents the Zookeeper state. Watch functions will receive a
    :class:`KeeperState` attribute as their state argument.

    .. attribute:: ASSOCIATING

        The Zookeeper ASSOCIATING state

    .. attribute:: AUTH_FAILED

        Authentication has failed, this is an unrecoverable error.

    .. attribute:: CONNECTED

        Zookeeper is connected.

    .. attribute:: CONNECTING

        Zookeeper is currently attempting to establish a connection.

    .. attribute:: EXPIRED_SESSION

        The prior session was invalid, all prior ephemeral nodes are
        gone.

    """
    ASSOCIATING = zookeeper.ASSOCIATING_STATE
    AUTH_FAILED = zookeeper.AUTH_FAILED_STATE
    CONNECTED = zookeeper.CONNECTED_STATE
    CONNECTING = zookeeper.CONNECTING_STATE
    EXPIRED_SESSION = zookeeper.EXPIRED_SESSION_STATE


class EventType(object):
    """Zookeeper Event

    Represents a Zookeeper event. Events trigger watch functions which
    will receive a :class:`EventType` attribute as their event
    argument.

    .. attribute:: NOTWATCHING

        This event type was added to Zookeeper in the event that
        watches get overloaded. It's never been used though and will
        likely be removed in a future Zookeeper version. **This event
        will never actually be set, don't bother testing for it.**

    .. attribute:: SESSION

        A Zookeeper session event. Watch functions do not receive
        session events. A session event watch can be registered with
        :class:`KazooClient` during creation that can receive these
        events. It's recommended to add a listener for connection state
        changes instead.

    .. attribute:: CREATED

        A node has been created.

    .. attribute:: DELETED

        A node has been deleted.

    .. attribute:: CHANGED

        The data for a node has changed.

    .. attribute:: CHILD

        The children under a node have changed (a child was added or
        removed). This event does not indicate the data for a child
        node has changed, which must have its own watch established.

    """
    NOTWATCHING = zookeeper.NOTWATCHING_EVENT
    SESSION = zookeeper.SESSION_EVENT
    CREATED = zookeeper.CREATED_EVENT
    DELETED = zookeeper.DELETED_EVENT
    CHANGED = zookeeper.CHANGED_EVENT
    CHILD = zookeeper.CHILD_EVENT


class WatchedEvent(namedtuple('WatchedEvent', ('type', 'state', 'path'))):
    """A change on ZooKeeper that a Watcher is able to respond to.

    The :class:`WatchedEvent` includes exactly what happened, the
    current state of ZooKeeper, and the path of the znode that was
    involved in the event. An instance of :class:`WatchedEvent` will be
    passed to registered watch functions.

    .. attribute:: type

        A :class:`EventType` attribute indicating the event type.

    .. attribute:: state

        A :class:`KeeperState` attribute indicating the Zookeeper
        state.

    .. attribute:: path

        The path of the node for the watch event.

    """


class ZnodeStat(namedtuple('ZnodeStat', ('aversion', 'ctime', 'cversion',
                                         'czxid', 'dataLength',
                                         'ephemeralOwner', 'mtime', 'mzxid',
                                         'numChildren', 'pzxid', 'version'))):
    """A ZnodeStat structure with convenience properties

    When getting the value of a node from Zookeeper, the properties for
    the node known as a "Stat structure" will be retrieved. The
    :class:`ZnodeStat` object provides access to the standard Stat
    properties and additional properties that are more readable and use
    Python time semantics (seconds since epoch instead of ms).

    .. note::

        The original Zookeeper Stat name is in parens next to the name
        when it differs from the convenience attribute. These are **not
        functions**, just attributes.

    .. attribute:: creation_transaction_id (czxid)

        The transaction id of the change that caused this znode to be
        created.

    .. attribute:: last_modified_transaction_id (mzxid)

        The transaction id of the change that last modified this znode.

    .. attribute:: created (ctime)

        The time in seconds from epoch when this node was created.
        (ctime is in milliseconds)

    .. attribute:: last_modified (mtime)

        The time in seconds from epoch when this znode was last
        modified. (mtime is in milliseconds)

    .. attribute:: version

        The number of changes to the data of this znode.

    .. attribute:: acl_version (aversion)

        The number of changes to the ACL of this znode.

    .. attribute:: owner_session_id (ephemeralOwner)

        The session id of the owner of this znode if the znode is an
        ephemeral node. If it is not an ephemeral node, it will be
        `None`. (ephemeralOwner will be 0 if it is not ephemeral)

    .. attribute:: data_length (dataLength)

        The length of the data field of this znode.

    .. attribute:: children_count (numChildren)

        The number of children of this znode.

    """
    @property
    def acl_version(self):
        return self.aversion

    @property
    def children_version(self):
        return self.cversion

    @property
    def created(self):
        return self.ctime / 1000.0

    @property
    def last_modified(self):
        return self.mtime / 1000.0

    @property
    def owner_session_id(self):
        return self.ephemeralOwner or None

    @property
    def creation_transaction_id(self):
        return self.czxid

    @property
    def last_modified_transaction_id(self):
        return self.mzxid

    @property
    def data_length(self):
        return self.dataLength

    @property
    def children_count(self):
        return self.numChildren


class Callback(namedtuple('Callback', ('type', 'func', 'args'))):
    """A callback that is handed to a handler for dispatch

    :param type: Type of the callback, can be 'session' or 'watch'
    :param func: Callback function
    :param args: Argument list for the callback function

    """


## Client Callbacks

def _generic_callback(async_result, handle, code, *args):
    if code != zookeeper.OK:
        exc = err_to_exception(code)
        async_result.set_exception(exc)
    else:
        if not args:
            result = None
        elif len(args) == 1:
            if isinstance(args[0], dict):
                # It's a node struct, put it in a ZnodeStat
                result = ZnodeStat(**args[0])
            else:
                result = args[0]
        else:
            # if there's two, the second is a stat object
            args = list(args)
            if len(args) == 2 and isinstance(args[1], dict):
                args[1] = ZnodeStat(**args[1])
            result = tuple(args)

        async_result.set(result)


def _exists_callback(async_result, handle, code, stat):
    if code not in (zookeeper.OK, zookeeper.NONODE):
        exc = err_to_exception(code)
        async_result.set_exception(exc)
    else:
        async_result.set(stat)


class KazooClient(object):
    """An Apache Zookeeper Python wrapper supporting alternate callback
    handlers and high-level functionality

    Watch functions registered with this class will not get session
    events, unlike the default Zookeeper watch's. They will also be
    called with a single argument, a :class:`WatchedEvent` instance.

    """
    # for testing purposes
    zookeeper = zookeeper

    def __init__(self, hosts='127.0.0.1:2181', watcher=None,
                 timeout=10.0, client_id=None, max_retries=None, retry_delay=0.1,
                 retry_backoff=2, retry_jitter=0.8, handler=None,
                 default_acl=None):
        """Create a KazooClient instance

        :param hosts: List of hosts to connect to
        :param watcher: Set a default watcher. This will be called by
                        the actual default watcher that
                        :class:`KazooClient` establishes.
        :param timeout: The longest to wait for a Zookeeper connection
        :param client_id: A Zookeeper client id, used when
                          re-establishing a prior session connection
        :param max_retries: Maximum retries when using the
                            :meth:`KazooClient.retry` method.
        :param retry_delay: Initial delay when retrying a call.
        :param retry_backoff: How much to back off on each call retry.
        :param retry_jitter: How much jitter delay to introduce per call. An
                             amount of time up to this will be added per retry
                             call to avoid hammering the server.
        :param handler: An instance of a class implementing the
                        :class:`~kazoo.interfaces.IHandler` interface
                        for callback handling
        :param default_acl: A default ACL used on node creation.
        """
        from kazoo.recipe.partitioner import SetPartitioner
        from kazoo.recipe.watchers import ChildrenWatch
        from kazoo.recipe.watchers import DataWatch

        # Check for chroot
        chroot_check = hosts.split('/', 1)
        if len(chroot_check) == 2:
            self.chroot = '/' + chroot_check[1]
        else:
            self.chroot = None

        # remove any trailing slashes
        self.default_acl = default_acl

        self._hosts = hosts
        self._watcher = watcher
        self._provided_client_id = client_id

        # ZK uses milliseconds
        self._timeout = int(timeout * 1000)

        # Record the handler strategy used
        self._handler = handler if handler else SequentialThreadingHandler()

        if inspect.isclass(self._handler):
            raise ConfigurationError("Handler must be an instance of a class, "
                                     "not the class: %s" % self._handler)

        self._handle = None

        # We use events like twitter's client to track current and desired
        # state (connected, and whether to shutdown)
        self._live = self._handler.event_object()
        self._stopped = self._handler.event_object()
        self._stopped.set()

        self._connection_timed_out = False

        self.retry = KazooRetry(
            max_tries=max_retries,
            delay=retry_delay,
            backoff=retry_backoff,
            max_jitter=retry_jitter,
            sleep_func=self._handler.sleep_func
        )

        # Curator like simplified state tracking, and listeners for state
        # transitions
        self.state = KazooState.LOST
        self.state_listeners = set()

        # convenience API
        self.ChildrenWatch = partial(ChildrenWatch, self)
        self.DataWatch = partial(DataWatch, self)
        self.Election = partial(Election, self)
        self.Lock = partial(Lock, self)
        self.Party = partial(Party, self)
        self.SetPartitioner = partial(SetPartitioner, self)
        self.ShallowParty = partial(ShallowParty, self)

    def _safe_call(self, func, async_result, *args, **kwargs):
        """Safely call a zookeeper function and handle errors related
        to a bad zhandle and state bugs

        In older zkpython bindings, a SystemError can arise if some
        functions are called when the session handle is expired. This
        client clears the old handle as soon as possible, but its
        possible a command may run against the old handle that is
        expired resulting in this error being thrown. See
        https://issues.apache.org/jira/browse/ZOOKEEPER-1318

        """
        try:
            return func(self._handle, *args)
        except BadArgumentsException:
            # Validate the path to throw a better except
            if isinstance(args[0], basestring) and not args[0].startswith('/'):
                raise ValueError("invalid path '%s'. must start with /" %
                                 args[0])
            else:
                raise
        except (TypeError, SystemError) as exc:
            # Handle was cleared or isn't set. If it was cleared and we are
            # supposed to be running, it means we had session expiration and
            # are attempting to reconnect. We treat it as a session expiration
            # in that case for appropriate behavior.
            if self._stopped.is_set():
                async_result.set_exception(
                    ZookeeperStoppedError(
                    "The Kazoo client is stopped, call connect before running "
                    "commands that talk to Zookeeper"))
            elif isinstance(exc, SystemError):
                # Set this to the error it should be for appropriate handling
                async_result.set_exception(
                    self.zookeeper.InvalidStateException(
                        "invalid handle state"))
            elif self._handle is None or 'an integer is required' in exc:
                async_result.set_exception(
                    self.zookeeper.SessionExpiredException("session expired"))
            else:
                raise

    def add_listener(self, listener):
        """Add a function to be called for connection state changes

        This function will be called with a :class:`KazooState`
        instance indicating the new connection state.

        """
        if not (listener and callable(listener)):
            raise ConfigurationError("listener must be callable")
        self.state_listeners.add(listener)

    def remove_listener(self, listener):
        """Remove a listener function"""
        self.state_listeners.discard(listener)

    @property
    def connected(self):
        """Returns whether the Zookeeper connection has been established"""
        return self._live.is_set()

    @property
    def client_id(self):
        """Returns the client id for this Zookeeper session if connected"""
        if self._handle is not None:
            return self.zookeeper.client_id(self._handle)
        return None

    def _wrap_session_callback(self, func):
        def wrapper(handle, type, state, path):
            callback = Callback('session', func, (handle, type, state, path))
            self._handler.dispatch_callback(callback)
        return wrapper

    def _wrap_watch_callback(self, func):
        def func_wrapper(*args):
            try:
                func(*args)
            except Exception:
                log.exception("Exception in kazoo callback <func: %s>" % func)

        def wrapper(handle, type, state, path):
            # don't send session events to all watchers
            if type != self.zookeeper.SESSION_EVENT:
                event = WatchedEvent(type, state, path)
                callback = Callback('watch', func_wrapper, (event,))
                self._handler.dispatch_callback(callback)
        return wrapper

    def _make_state_change(self, state):
        # skip if state is current
        if self.state == state:
            return
        self.state = state

        # Create copy of listeners for iteration in case one needs to
        # remove itself
        for listener in list(self.state_listeners):
            try:
                listener(state)
            except Exception:
                log.exception("Error in connection state listener")

    def _session_callback(self, handle, type, state, path):
        if type != EventType.SESSION:
            return

        if self._handle != handle:
            try:
                # latent handle callback from previous connection
                self.zookeeper.close(handle)
            except Exception:
                pass
            return

        if self._stopped.is_set():
            # Any events at this point can be ignored
            return

        if state == KeeperState.CONNECTED:
            self._live.set()
            self._make_state_change(KazooState.CONNECTED)
        elif state in (KeeperState.EXPIRED_SESSION,
                       KeeperState.AUTH_FAILED):
            self._live.clear()
            self._make_state_change(KazooState.LOST)
            self._handle = None
            self.connect_async()
        else:
            # Connection lost
            self._live.clear()
            self._make_state_change(KazooState.SUSPENDED)

        if self._watcher:
            self._watcher(WatchedEvent(type, state, path))

    def _safe_close(self):
        if self._handle is not None:
            zh, self._handle = self._handle, None
            try:
                self.zookeeper.close(zh)
            except self.zookeeper.ZooKeeperException:
                # Corrupt session or otherwise disconnected
                pass
            self._live.clear()
            self._make_state_change(KazooState.LOST)

    def connect_async(self):
        """Asynchronously initiate connection to ZK

        :returns: An AsyncResult instance for the configured handler
        :rtype: :class:`~kazoo.interfaces.IAsyncResult`

        """
        # If we're already connected, ignore
        if self._live.is_set():
            return

        # Make sure we're safely closed
        self._safe_close()

        # We've been asked to connect, clear the stop
        self._stopped.clear()

        cb = self._wrap_session_callback(self._session_callback)
        if self._provided_client_id:
            # TODO: this doesn't seem to account for self.namespace?!
            self._handle = self.zookeeper.init(self._hosts, cb, self._timeout,
                self._provided_client_id)
        else:
            self._handle = self.zookeeper.init(self._hosts, cb, self._timeout)
        return self._live

    def connect(self, timeout=15):
        """Initiate connection to ZK

        :param timeout: Time in seconds to wait for connection to
                        succeed.

        """
        event = self.connect_async()

        if not event:
            # Already connected
            return

        event.wait(timeout=timeout)
        if not self.connected:
            # We time-out, ensure we are disconnected
            self.stop()
            raise self._handler.timeout_exception("Connection time-out")

    def stop(self):
        """Gracefully stop this Zookeeper session"""
        self._stopped.set()
        self._safe_close()

    def restart(self):
        """Stop and restart the Zookeeper session."""
        self._safe_close()
        self._stopped.clear()
        self.connect()

    def add_auth_async(self, scheme, credential):
        """Asynchronously send credentials to server

        :param scheme: authentication scheme (default supported:
                       "digest")
        :param credential: the credential -- value depends on scheme
        :returns: AsyncResult object set on completion
        :rtype: :class:`~kazoo.interfaces.IAsyncResult`

        """
        async_result = self._handler.async_result()
        callback = partial(_generic_callback, async_result)

        self._safe_call(self.zookeeper.add_auth, async_result, scheme,
                        credential, callback)

        # Compensate for io polling bug on auth by running an exists call
        # See https://issues.apache.org/jira/browse/ZOOKEEPER-770
        self.exists_async("/")
        return async_result

    def add_auth(self, scheme, credential):
        """Send credentials to server

        :param scheme: authentication scheme (default supported:
                       "digest")
        :param credential: the credential -- value depends on scheme

        """
        return self.add_auth_async(scheme, credential).get()

    def ensure_path(self, path, acl=None):
        """Recursively create a path if it doesn't exist
        """
        self._inner_ensure_path(path, acl)

    def _inner_ensure_path(self, path, acl):
        if self.exists(path):
            return

        if acl is None and self.default_acl:
            acl = self.default_acl

        parent, node = split(path)

        if node:
            self._inner_ensure_path(parent, acl)
        try:
            self.create_async(path, "", acl=acl).get()
        except self.zookeeper.NodeExistsException:
            # someone else created the node. how sweet!
            pass

    def unchroot(self, path):
        if not self.chroot:
            return path

        if path.startswith(self.chroot):
            return path[len(self.chroot):]
        else:
            return path

    def create_async(self, path, value, acl=None, ephemeral=False,
                     sequence=False):
        """Asynchronously create a ZNode

        :param path: path of node
        :param value: initial value of node
        :param acl: permissions for node
        :param ephemeral: boolean indicating whether node is ephemeral
                          (tied to this session)
        :param sequence: boolean indicating whether path is suffixed
                         with a unique index
        :returns: AsyncResult object set on completion with the real
                  path of the new node
        :rtype: :class:`~kazoo.interfaces.IAsyncResult`

        """
        if acl is None and self.default_acl:
            acl = self.default_acl

        flags = 0
        if ephemeral:
            flags |= self.zookeeper.EPHEMERAL
        if sequence:
            flags |= self.zookeeper.SEQUENCE
        if acl is None:
            acl = (ZK_OPEN_ACL_UNSAFE,)

        async_result = self._handler.async_result()
        callback = partial(_generic_callback, async_result)

        self._safe_call(self.zookeeper.acreate, async_result, path, value,
                        list(acl), flags, callback)
        return async_result

    def create(self, path, value, acl=None, ephemeral=False, sequence=False,
               makepath=False):
        """Create a ZNode

        :param path: path of node
        :param value: initial value of node
        :param acl: permissions for node
        :param ephemeral: boolean indicating whether node is ephemeral
                          (tied to this session)
        :param sequence: boolean indicating whether path is suffixed
                         with a unique index
        :param makepath: Whether the path should be created if it
                         doesn't exist
        :returns: real path of the new node

        """
        try:
            realpath = self.create_async(path, value, acl=acl,
                ephemeral=ephemeral, sequence=sequence).get()

        except NoNodeException:
            # some or all of the parent path doesn't exist. if makepath is set
            # we will create it and retry. If it fails again, someone must be
            # actively deleting ZNodes and we'd best bail out.
            if not makepath:
                raise

            parent, _ = split(path)

            # using the inner call directly because path is already namespaced
            self._inner_ensure_path(parent, acl)

            # now retry
            realpath = self.create_async(path, value, acl=acl,
                ephemeral=ephemeral, sequence=sequence).get()

        return self.unchroot(realpath)

    def exists_async(self, path, watch=None):
        """Asynchronously check if a node exists

        :param path: path of node
        :param watch: optional watch callback to set for future changes
                      to this path
        :returns: stat of the node if it exists, else None
        :rtype: `dict` or `None`

        """
        async_result = self._handler.async_result()
        callback = partial(_exists_callback, async_result)
        watch_callback = self._wrap_watch_callback(watch) if watch else None

        self._safe_call(self.zookeeper.aexists, async_result, path,
                        watch_callback, callback)
        return async_result

    def exists(self, path, watch=None):
        """Check if a node exists

        :param path: path of node
        :param watch: optional watch callback to set for future changes
                      to this path
        :returns: stat of the node if it exists, else None
        :rtype: `dict` or `None`

        """
        return self.exists_async(path, watch).get()

    def get_async(self, path, watch=None):
        """Asynchronously get the value of a node

        :param path: path of node
        :param watch: optional watch callback to set for future changes
                      to this path
        :returns: AsyncResult set with tuple (value, :class:`ZnodeStat`
                  ) of node on success
        :rtype: :class:`~kazoo.interfaces.IAsyncResult`

        """
        async_result = self._handler.async_result()
        callback = partial(_generic_callback, async_result)
        watch_callback = self._wrap_watch_callback(watch) if watch else None

        self._safe_call(self.zookeeper.aget, async_result, path,
                        watch_callback, callback)
        return async_result

    def get(self, path, watch=None):
        """Get the value of a node

        :param path: path of node
        :param watch: optional watch callback to set for future changes
                      to this path
        :returns: tuple (value, :class:`ZnodeStat`) of node

        """
        return self.get_async(path, watch).get()

    def get_children_async(self, path, watch=None):
        """Asynchronously get a list of child nodes of a path

        :param path: path of node to list
        :param watch: optional watch callback to set for future changes
                      to this path
        :returns: AsyncResult set with list of child node names on
                  success
        :rtype: :class:`~kazoo.interfaces.IAsyncResult`

        """
        async_result = self._handler.async_result()
        callback = partial(_generic_callback, async_result)
        watch_callback = self._wrap_watch_callback(watch) if watch else None

        self._safe_call(self.zookeeper.aget_children, async_result, path,
                        watch_callback, callback)
        return async_result

    def get_children(self, path, watch=None):
        """Get a list of child nodes of a path

        :param path: path of node to list
        :param watch: optional watch callback to set for future changes
                      to this path
        :returns: list of child node names

        """
        return self.get_children_async(path, watch).get()

    def set_async(self, path, data, version=-1):
        """Set the value of a node

        If the version of the node being updated is newer than the
        supplied version (and the supplied version is not -1), a
        BadVersionException will be raised.

        :param path: path of node to set
        :param data: new data value
        :param version: version of node being updated, or -1
        :returns: AsyncResult set with new node :class:`ZnodeStat` on
                  success
        :rtype: :class:`~kazoo.interfaces.IAsyncResult`

        """
        async_result = self._handler.async_result()
        callback = partial(_generic_callback, async_result)

        self._safe_call(self.zookeeper.aset, async_result, path, data, version,
                        callback)
        return async_result

    def set(self, path, data, version=-1):
        """Set the value of a node

        If the version of the node being updated is newer than the
        supplied version (and the supplied version is not -1), a
        BadVersionException will be raised.

        :param path: path of node to set
        :param data: new data value
        :param version: version of node being updated, or -1
        :returns: updated :class:`ZnodeStat` of the node

        """
        return self.set_async(path, data, version).get()

    def delete_async(self, path, version=-1):
        """Asynchronously delete a node

        :param path: path of node to delete
        :param version: version of node to delete, or -1 for any
        :returns: AyncResult set upon completion
        :rtype: :class:`~kazoo.interfaces.IAsyncResult`
        """
        async_result = self._handler.async_result()
        callback = partial(_generic_callback, async_result)

        self._safe_call(self.zookeeper.adelete, async_result, path, version,
                        callback)
        return async_result

    def delete(self, path, version=-1, recursive=False):
        """Delete a node

        :param path: path of node to delete
        :param version: version of node to delete, or -1 for any
        :param recursive: Recursively delete node and all its children,
            defaults to False.
        """
        if recursive:
            self._delete_recursive(path)
        else:
            self.delete_async(path, version).get()

    def _delete_recursive(self, path):
        try:
            children = self.get_children(path)
        except self.zookeeper.NoNodeException:
            return

        if children:
            for child in children:
                if path == "/":
                    child_path = path + child
                else:
                    child_path = path + "/" + child

                self._delete_recursive(child_path)
        try:
            self.delete(path)
        except self.zookeeper.NoNodeException:
            pass



import logging
import os
import threading
import uuid
import unittest

import zookeeper
from nose.tools import eq_

import kazoo.client
from kazoo.testing import KazooTestCase
from kazoo.testing import ZooError
from kazoo.exceptions import NoNodeException
from kazoo.exceptions import NoAuthException
from kazoo.tests.util import InstalledHandler
from kazoo.tests.util import wait


class LoggingTests(unittest.TestCase):
    def test_logging(self):
        handler = InstalledHandler('ZooKeeper')
        try:
            handle = zookeeper.init('zookeeper.example.com:2181')
            zookeeper.close(handle)
        except Exception:
            pass

        wait(lambda: [r for r in handler.records
                       if 'environment' in r.getMessage()]
             )
        handler.clear()

        # Test that the filter for the "Exceeded deadline by" noise works.
        # cheat and bypass zk by writing to the pipe directly.
        os.write(kazoo.client._logging_pipe[1],
                 '2012-01-06 16:45:44,572:43673(0x1004f6000):ZOO_WARN@'
                 'zookeeper_interest@1461: Exceeded deadline by 27747ms\n')
        wait(lambda: [r for r in handler.records
                       if ('Exceeded deadline by' in r.getMessage()
                           and r.levelno == logging.DEBUG)
                       ]
             )

        self.assertFalse([r for r in handler.records
                          if ('Exceeded deadline by' in r.getMessage()
                              and r.levelno == logging.WARNING)
                          ])

        handler.uninstall()


class TestConnection(KazooTestCase):
    def _makeAuth(self, *args, **kwargs):
        from kazoo.security import make_digest_acl
        return make_digest_acl(*args, **kwargs)

    def test_auth(self):
        username = uuid.uuid4().hex
        password = uuid.uuid4().hex

        digest_auth = "%s:%s" % (username, password)
        acl = self._makeAuth(username, password, all=True)

        self.client.add_auth("digest", digest_auth)

        self.client.default_acl = (acl,)

        self.client.create("/1", "")
        self.client.create("/1/2", "")

        eve = self._get_client()
        eve.connect()

        try:
            self.assertRaises(NoAuthException, eve.get, "/1/2")

            # try again with the wrong auth token
            eve.add_auth("digest", "badbad:bad")

            self.assertRaises(NoAuthException, eve.get, "/1/2")
        finally:
            # Ensure we remove the ACL protected nodes
            self.client.delete("/1", recursive=True)

    def test_session_expire(self):
        from kazoo.client import KazooState

        cv = threading.Event()

        def watch_events(event):
            if event == KazooState.LOST:
                cv.set()

        self.client.add_listener(watch_events)
        self.expire_session()
        cv.wait(1)
        assert cv.is_set()

    def test_bad_session_expire(self):
        from kazoo.client import KazooState

        cv = threading.Event()

        def watch_events(event):
            if event == KazooState.LOST:
                raise Exception("oops")
                cv.set()

        self.client.add_listener(watch_events)
        self.expire_session()
        cv.wait(0.5)
        assert not cv.is_set()

    def test_state_listener(self):
        from kazoo.client import KazooState
        states = []
        condition = threading.Condition()

        def listener(state):
            with condition:
                states.append(state)
                condition.notify_all()

        self.client.stop()
        self.client.add_listener(listener)
        self.client.connect(5)

        with condition:
            if not states:
                condition.wait(5)

        eq_(len(states), 1)
        eq_(states[0], KazooState.CONNECTED)

    def test_no_connection(self):
        from kazoo.exceptions import ZookeeperStoppedError
        self.client.stop()
        self.assertRaises(ZookeeperStoppedError, self.client.exists, '/')


class TestClient(KazooTestCase):
    def _getKazooState(self):
        from kazoo.client import KazooState
        return KazooState

    def test_legacy_error(self):
        self.add_errors(dict(
            acreate=[ZooError('call', SystemError, False)]
        ))
        self.assertRaises(zookeeper.InvalidStateException, self.client.create,
                          "/1", "val1")

    def test_bad_arguments(self):
        self.add_errors(dict(
            acreate=[ZooError('call', zookeeper.BadArgumentsException, False)]
        ))
        self.assertRaises(zookeeper.BadArgumentsException, self.client.create,
                          "/1", "val1")

    def test_bad_handle_type_error(self):
        self.add_errors(dict(
            acreate=[ZooError('call', TypeError("an integer is required"),
                              False)]
        ))
        self.assertRaises(zookeeper.SessionExpiredException, self.client.create,
                          "/1", "val1")

    def test_bad_argument(self):
        client = self.client
        client.ensure_path("/1")
        self.assertRaises(TypeError, self.client.set, "/1", 1)

    def test_ensure_path(self):
        client = self.client
        client.ensure_path("/1/2")
        self.assertTrue(client.exists("/1/2"))

        client.ensure_path("/1/2/3/4")
        self.assertTrue(client.exists("/1/2/3/4"))

    def test_create_no_makepath(self):
        self.assertRaises(NoNodeException, self.client.create, "/1/2", "val1")
        self.assertRaises(NoNodeException, self.client.create, "/1/2", "val1",
            makepath=False)

    def test_bad_create_args(self):
        # We need a non-namespaced client for this test
        client = self._get_nonchroot_client()
        try:
            client.connect()
            self.assertRaises(ValueError, client.create, "1/2", "val1")
        finally:
            client.stop()

    def test_create_makepath(self):
        self.client.create("/1/2", "val1", makepath=True)
        data, stat = self.client.get("/1/2")
        eq_(data, "val1")

        self.client.create("/1/2/3/4/5", "val2", makepath=True)
        data, stat = self.client.get("/1/2/3/4/5")
        eq_(data, "val2")

    def test_create_get_set(self):
        nodepath = "/" + uuid.uuid4().hex

        self.client.create(nodepath, "sandwich", ephemeral=True)

        data, stat = self.client.get(nodepath)
        eq_(data, "sandwich")

        newstat = self.client.set(nodepath, "hats", stat.version)
        self.assertTrue(newstat)
        assert newstat.version > stat.version

        # Some other checks of the ZnodeStat object we got
        eq_(newstat.acl_version, stat.acl_version)
        eq_(newstat.created, stat.ctime / 1000.0)
        eq_(newstat.last_modified, newstat.mtime / 1000.0)
        eq_(newstat.owner_session_id, stat.ephemeralOwner)
        eq_(newstat.creation_transaction_id, stat.czxid)
        eq_(newstat.last_modified_transaction_id, newstat.mzxid)
        eq_(newstat.data_length, newstat.dataLength)
        eq_(newstat.children_count, stat.numChildren)
        eq_(newstat.children_version, stat.cversion)

    def test_create_get_sequential(self):
        basepath = "/" + uuid.uuid4().hex
        realpath = self.client.create(basepath, "sandwich", sequence=True,
            ephemeral=True)

        self.assertTrue(basepath != realpath and realpath.startswith(basepath))

        data, stat = self.client.get(realpath)
        eq_(data, "sandwich")

    def test_exists(self):
        nodepath = "/" + uuid.uuid4().hex

        exists = self.client.exists(nodepath)
        eq_(exists, None)

        self.client.create(nodepath, "sandwich", ephemeral=True)
        exists = self.client.exists(nodepath)
        self.assertTrue(exists)
        assert "version" in exists

        multi_node_nonexistent = "/" + uuid.uuid4().hex + "/hats"
        exists = self.client.exists(multi_node_nonexistent)
        eq_(exists, None)

    def test_exists_watch(self):
        nodepath = "/" + uuid.uuid4().hex

        event = threading.Event()

        def w(watch_event):
            eq_(watch_event.path, nodepath)
            event.set()

        exists = self.client.exists(nodepath, watch=w)
        eq_(exists, None)

        self.client.create(nodepath, "x", ephemeral=True)

        event.wait(1)
        self.assertTrue(event.is_set())

    def test_exists_watcher_exception(self):
        nodepath = "/" + uuid.uuid4().hex

        event = threading.Event()

        # if the watcher throws an exception, all we can really do is log it
        def w(watch_event):
            eq_(watch_event.path, nodepath)
            event.set()

            raise Exception("test exception in callback")

        exists = self.client.exists(nodepath, watch=w)
        eq_(exists, None)

        self.client.create(nodepath, "x", ephemeral=True)

        event.wait(1)
        self.assertTrue(event.is_set())

    def test_create_delete(self):
        nodepath = "/" + uuid.uuid4().hex

        self.client.create(nodepath, "zzz")

        self.client.delete(nodepath)

        exists = self.client.exists(nodepath)
        eq_(exists, None)

dummy_dict = {
    'aversion': 1, 'ctime': 0, 'cversion': 1,
    'czxid': 110, 'dataLength': 1, 'ephemeralOwner': 'ben',
    'mtime': 1, 'mzxid': 1, 'numChildren': 0, 'pzxid': 1, 'version': 1
}


class TestCallbacks(unittest.TestCase):
    def test_exists_callback(self):
        from kazoo.client import _exists_callback
        from kazoo.handlers.threading import SequentialThreadingHandler
        handler = SequentialThreadingHandler()
        asy = handler.async_result()
        _exists_callback(asy, 0, zookeeper.OK, True)
        eq_(asy.get(), True)

        asy = handler.async_result()
        _exists_callback(asy, 0, zookeeper.CONNECTIONLOSS, False)
        self.assertRaises(zookeeper.ConnectionLossException, asy.get)

    def test_generic_callback_ok(self):
        from kazoo.client import _generic_callback
        from kazoo.handlers.threading import SequentialThreadingHandler
        handler = SequentialThreadingHandler()

        # No args
        asy = handler.async_result()
        _generic_callback(asy, 0, zookeeper.OK)
        eq_(asy.get(), None)

        # One arg thats not a dict
        asy = handler.async_result()
        _generic_callback(asy, 0, zookeeper.OK, 12)
        eq_(asy.get(), 12)

        # One arg thats a node struct
        asy = handler.async_result()
        _generic_callback(asy, 0, zookeeper.OK, dummy_dict)
        eq_(asy.get().acl_version, 1)

        # two args, second is struct
        asy = handler.async_result()
        _generic_callback(asy, 0, zookeeper.OK, 11, dummy_dict)
        val = asy.get()
        eq_(val[1].acl_version, 1)
        eq_(val[0], 11)

    def test_generic_callback_error(self):
        from kazoo.client import _generic_callback
        from kazoo.handlers.threading import SequentialThreadingHandler
        handler = SequentialThreadingHandler()

        asy = handler.async_result()
        _generic_callback(asy, 0, zookeeper.CONNECTIONLOSS)
        self.assertRaises(zookeeper.ConnectionLossException, asy.get)

    def test_session_callback_states(self):
        from kazoo.client import (KazooClient, KazooState, KeeperState,
            EventType)

        client = KazooClient()
        client._handle = 1
        client._live.set()

        result = client._session_callback(1, EventType.CREATED,
                                          KeeperState.CONNECTED, '/')
        eq_(result, None)

        # Now with stopped
        client._stopped.set()
        result = client._session_callback(1, EventType.SESSION,
                                          KeeperState.CONNECTED, '/')
        eq_(result, None)

        # Test several state transitions
        client._stopped.clear()
        client.connect_async = lambda: True
        client._session_callback(1, EventType.SESSION, KeeperState.CONNECTED,
                                 None)
        eq_(client.state, KazooState.CONNECTED)

        client._session_callback(1, EventType.SESSION, KeeperState.AUTH_FAILED,
                                 None)
        eq_(client._handle, None)
        eq_(client.state, KazooState.LOST)

        client._handle = 1
        client._session_callback(1, EventType.SESSION, -250, None)
        eq_(client.state, KazooState.SUSPENDED)

        # handle mismatch
        client._handle = 0
        # This will be ignored due to handle mismatch
        client._session_callback(1, EventType.SESSION, KeeperState.CONNECTED,
                                 None)
        eq_(client.state, KazooState.SUSPENDED)



import random
import time

from zookeeper import (
    ClosingException,
    ConnectionLossException,
    InvalidStateException,
    OperationTimeoutException,
    SessionExpiredException
)


class ForceRetryError(Exception):
    """Raised when some recipe logic wants to force a retry"""


class KazooRetry(object):
    """Helper for retrying a method in the face of retry-able exceptions"""
    RETRY_EXCEPTIONS = (
        ClosingException,
        ConnectionLossException,
        OperationTimeoutException,
        ForceRetryError
    )

    EXPIRED_EXCEPTIONS = (
        SessionExpiredException,

        # Occurs when a command is run on a session handle that expired, if it
        # manages to run exactly when it expired but before the handle was
        # removed for reconnection
        InvalidStateException,
    )

    def __init__(self, max_tries=1, delay=0.1, backoff=2, max_jitter=0.8,
                 ignore_expire=True, sleep_func=time.sleep):
        """Create a :class:`KazooRetry` instance

        :param max_tries: How many times to retry the command.
        :param delay: Initial delay between retry attempts
        :param backoff: Backoff multiplier between retry attempts. Defaults
                        to 2 for exponential backoff.
        :param max_jitter: Additional max jitter period to wait between retry
                           attempts to avoid slamming the server.
        :param ignore_expire: Whether a session expiration should be ignored
                              and treated as a retry-able command.

        """
        self.sleep_func = sleep_func
        self.max_tries = max_tries
        self.delay = delay
        self.backoff = backoff
        self.max_jitter = int(max_jitter * 100)
        self.retry_exceptions = self.RETRY_EXCEPTIONS
        if ignore_expire:
            self.retry_exceptions += self.EXPIRED_EXCEPTIONS

    def run(self, func, *args, **kwargs):
        self(func, *args, **kwargs)

    def __call__(self, func, *args, **kwargs):
        tries = 1
        delay = self.delay

        while True:
            try:
                return func(*args, **kwargs)

            except self.retry_exceptions:
                if tries == self.max_tries:
                    raise
                tries += 1
                jitter = random.randint(0, self.max_jitter) / 100.0
                self.sleep_func(delay + jitter)
                delay *= self.backoff



"""Zookeeper exceptions and mappings"""
import zookeeper
from zookeeper import (
    SystemErrorException,
    RuntimeInconsistencyException,
    DataInconsistencyException,
    ConnectionLossException,
    MarshallingErrorException,
    UnimplementedException,
    OperationTimeoutException,
    BadArgumentsException,
    ApiErrorException,
    NoNodeException,
    NoAuthException,
    BadVersionException,
    NoChildrenForEphemeralsException,
    NodeExistsException,
    InvalidACLException,
    AuthFailedException,
    NotEmptyException,
    SessionExpiredException,
    InvalidCallbackException,
    InvalidStateException
)


class KazooException(Exception):
    """Base Kazoo exception that all other kazoo library exceptions inherit
    from"""


class CancelledError(KazooException):
    """Raised when a process is cancelled by another thread"""


class ConfigurationError(KazooException):
    """Raised if the configuration arguments to an object are invalid"""


class ZookeeperStoppedError(KazooException):
    """Raised when the kazoo client stopped (and thus not connected)"""


# this dictionary is a port of err_to_exception() from zkpython zookeeper.c
_ERR_TO_EXCEPTION = {
    zookeeper.SYSTEMERROR: SystemErrorException,
    zookeeper.RUNTIMEINCONSISTENCY: RuntimeInconsistencyException,
    zookeeper.DATAINCONSISTENCY: DataInconsistencyException,
    zookeeper.CONNECTIONLOSS: ConnectionLossException,
    zookeeper.MARSHALLINGERROR: MarshallingErrorException,
    zookeeper.UNIMPLEMENTED: UnimplementedException,
    zookeeper.OPERATIONTIMEOUT: OperationTimeoutException,
    zookeeper.BADARGUMENTS: BadArgumentsException,
    zookeeper.APIERROR: ApiErrorException,
    zookeeper.NONODE: NoNodeException,
    zookeeper.NOAUTH: NoAuthException,
    zookeeper.BADVERSION: BadVersionException,
    zookeeper.NOCHILDRENFOREPHEMERALS: NoChildrenForEphemeralsException,
    zookeeper.NODEEXISTS: NodeExistsException,
    zookeeper.INVALIDACL: InvalidACLException,
    zookeeper.AUTHFAILED: AuthFailedException,
    zookeeper.NOTEMPTY: NotEmptyException,
    zookeeper.SESSIONEXPIRED: SessionExpiredException,
    zookeeper.INVALIDCALLBACK: InvalidCallbackException,
    zookeeper.INVALIDSTATE: InvalidStateException,
}


def err_to_exception(error_code, msg=None):
    """Return an exception object for a Zookeeper error code
    """
    try:
        zkmsg = zookeeper.zerror(error_code)
    except Exception:
        zkmsg = ""

    if msg:
        if zkmsg:
            msg = "%s: %s" % (zkmsg, msg)
    else:
        msg = zkmsg

    exc = _ERR_TO_EXCEPTION.get(error_code)
    if exc is None:

        # double check that it isn't an ok resonse
        if error_code == zookeeper.OK:
            return None

        # otherwise generic exception
        exc = Exception
    return exc(msg)



import threading
import unittest

import mock
from nose.tools import eq_
from nose.tools import raises

from kazoo.client import Callback


class TestThreadingHandler(unittest.TestCase):
    def _makeOne(self, *args):
        from kazoo.handlers.threading import SequentialThreadingHandler
        return SequentialThreadingHandler(*args)

    def _getAsync(self, *args):
        from kazoo.handlers.threading import AsyncResult
        return AsyncResult

    def test_completion_vs_session(self):
        h = self._makeOne()

        lst = []
        av = threading.Event()
        bv = threading.Event()
        cv = threading.Event()
        dv = threading.Event()

        def addToList():
            av.set()
            bv.wait()
            lst.append(True)
            dv.set()

        def anotherAdd():
            lst.append(True)
            cv.set()

        call1 = Callback('session', addToList, ())
        call2 = Callback('completion', anotherAdd, ())

        h.dispatch_callback(call1)
        av.wait()
        # Now we know the first is waiting, make sure
        # the second executes while the first has blocked
        # its thread
        h.dispatch_callback(call2)
        cv.wait()

        eq_(lst, [True])
        bv.set()
        dv.wait()
        eq_(lst, [True, True])

    def test_proper_threading(self):
        h = self._makeOne()
        assert isinstance(h.event_object(), threading._Event)

    def test_matching_async(self):
        h = self._makeOne()
        async = self._getAsync()
        assert isinstance(h.async_result(), async)


class TestThreadingAsync(unittest.TestCase):
    def _makeOne(self, *args):
        from kazoo.handlers.threading import AsyncResult
        return AsyncResult(*args)

    def _makeHandler(self):
        from kazoo.handlers.threading import SequentialThreadingHandler
        return SequentialThreadingHandler()

    def test_ready(self):
        mock_handler = mock.Mock()
        async = self._makeOne(mock_handler)

        eq_(async.ready(), False)
        async.set('val')
        eq_(async.ready(), True)
        eq_(async.successful(), True)
        eq_(async.exception, None)

    def test_callback_queued(self):
        mock_handler = mock.Mock()
        mock_handler.completion_queue = mock.Mock()
        async = self._makeOne(mock_handler)

        async.rawlink(lambda a: a)
        async.set('val')

        assert mock_handler.completion_queue.put.called

    def test_set_exception(self):
        mock_handler = mock.Mock()
        mock_handler.completion_queue = mock.Mock()
        async = self._makeOne(mock_handler)
        async.rawlink(lambda a: a)
        async.set_exception(ImportError('Error occured'))

        assert isinstance(async.exception, ImportError)
        assert mock_handler.completion_queue.put.called

    def test_get_wait_while_setting(self):
        mock_handler = mock.Mock()
        async = self._makeOne(mock_handler)

        lst = []
        bv = threading.Event()
        cv = threading.Event()

        def wait_for_val():
            bv.set()
            val = async.get()
            lst.append(val)
            cv.set()
        th = threading.Thread(target=wait_for_val)
        th.start()
        bv.wait()

        async.set('fred')
        cv.wait()
        eq_(lst, ['fred'])

    def test_get_with_nowait(self):
        mock_handler = mock.Mock()
        async = self._makeOne(mock_handler)
        timeout = self._makeHandler().timeout_exception

        @raises(timeout)
        def test_it():
            async.get(block=False)
        test_it()

        @raises(timeout)
        def test_nowait():
            async.get_nowait()
        test_nowait()

    def test_get_with_exception(self):
        mock_handler = mock.Mock()
        async = self._makeOne(mock_handler)

        lst = []
        bv = threading.Event()
        cv = threading.Event()

        def wait_for_val():
            bv.set()
            try:
                val = async.get()
            except ImportError:
                lst.append('oops')
            else:
                lst.append(val)
            cv.set()
        th = threading.Thread(target=wait_for_val)
        th.start()
        bv.wait()

        async.set_exception(ImportError)
        cv.wait()
        eq_(lst, ['oops'])

    def test_set_before_wait(self):
        mock_handler = mock.Mock()
        async = self._makeOne(mock_handler)

        lst = []
        cv = threading.Event()
        async.set('fred')

        def wait_for_val():
            val = async.get()
            lst.append(val)
            cv.set()
        th = threading.Thread(target=wait_for_val)
        th.start()
        cv.wait()
        eq_(lst, ['fred'])

    def test_set_exc_before_wait(self):
        mock_handler = mock.Mock()
        async = self._makeOne(mock_handler)

        lst = []
        cv = threading.Event()
        async.set_exception(ImportError)

        def wait_for_val():
            try:
                val = async.get()
            except ImportError:
                lst.append('ooops')
            else:
                lst.append(val)
            cv.set()
        th = threading.Thread(target=wait_for_val)
        th.start()
        cv.wait()
        eq_(lst, ['ooops'])

    def test_linkage(self):
        mock_handler = mock.Mock()
        async = self._makeOne(mock_handler)
        cv = threading.Event()

        lst = []

        def add_on():
            lst.append(True)

        def wait_for_val():
            async.get()
            cv.set()

        th = threading.Thread(target=wait_for_val)
        th.start()

        async.rawlink(add_on)
        async.set('fred')
        assert mock_handler.completion_queue.put.called
        async.unlink(add_on)
        cv.wait()
        eq_(async.value, 'fred')

    def test_linkage_not_ready(self):
        mock_handler = mock.Mock()
        async = self._makeOne(mock_handler)

        lst = []

        def add_on():
            lst.append(True)

        async.set('fred')
        assert not mock_handler.completion_queue.called
        async.rawlink(add_on)
        assert mock_handler.completion_queue.put.called

    def test_link_and_unlink(self):
        mock_handler = mock.Mock()
        async = self._makeOne(mock_handler)

        lst = []

        def add_on():
            lst.append(True)

        async.rawlink(add_on)
        assert not mock_handler.completion_queue.put.called
        async.unlink(add_on)
        async.set('fred')
        assert not mock_handler.completion_queue.put.called



"""gevent Handler"""
from __future__ import absolute_import

import fcntl
import os

import gevent
import gevent.event
import gevent.thread
try:
    from gevent import get_hub
except ImportError:  # pragma: nocover
    from gevent.hub import get_hub

from gevent.queue import Empty
from gevent.queue import Queue
from zope.interface import implementer

from kazoo.interfaces import IAsyncResult
from kazoo.interfaces import IHandler


if gevent.__version__.startswith('1.'):
    _using_libevent = False
else:
    _using_libevent = True


# Simple wrapper to os.pipe() - but sets to non-block
def _pipe():
    r, w = os.pipe()
    fcntl.fcntl(r, fcntl.F_SETFL, os.O_NONBLOCK)
    fcntl.fcntl(w, fcntl.F_SETFL, os.O_NONBLOCK)
    return r, w

_core_pipe_read, _core_pipe_write = _pipe()


def _core_pipe_read_callback(event, evtype):
    try:
        os.read(event.fd, 1)
    except Exception:
        pass

if _using_libevent:
    # The os pipe trick to wake gevent is only need for gevent pre-1.0
    gevent.core.event(gevent.core.EV_READ | gevent.core.EV_PERSIST,
                     _core_pipe_read, _core_pipe_read_callback).add()


@implementer(IAsyncResult)
class AsyncResult(gevent.event.AsyncResult):
    """A gevent AsyncResult capable of waking the gevent thread when used
    from the Zookeeper thread"""
    def __init__(self, handler):
        self._handler = handler
        gevent.event.AsyncResult.__init__(self)

    def set(self, value=None):
        # Proxy the set call to the gevent thread
        self._handler.completion_queue.put(
            lambda: gevent.event.AsyncResult.set(self, value)
        )
        self._handler.wake()

    def set_exception(self, exception):
        # Proxy the set_exception call to the gevent thread
        self._handler.completion_queue.put(
            lambda: gevent.event.AsyncResult.set_exception(self, exception)
        )
        self._handler.wake()


@implementer(IHandler)
class SequentialGeventHandler(object):
    """gevent Handler for sequentially executing callbacks

    This handler executes callbacks in a sequential manner from the Zookeeper
    thread. A queue is created for each of the callback events, so that each
    type of event has its callback type run sequentially. These are split into
    three queues, therefore it's possible that a session event arriving after a
    watch event may have its callback executed at the same time or slightly
    before the watch event callback.

    Each queue type has a greenlet worker that pulls the callback event off the
    queue and runs it in the order Zookeeper sent it.

    This split helps ensure that watch callbacks won't block session
    re-establishment should the connection be lost during a Zookeeper client
    call.

    Watch callbacks and session callbacks should avoid blocking behavior as the
    next callback of that type won't be run until it completes. If you need
    to block, spawn a new greenlet and return immediately so callbacks can
    proceed.

    """
    name = "sequential_gevent_handler"
    timeout_exception = gevent.event.Timeout
    sleep_func = gevent.sleep

    def __init__(self, hub=None):
        """Create a :class:`SequentialGeventHandler` instance"""
        self.completion_queue = Queue()
        self.callback_queue = Queue()
        self.session_queue = Queue()
        self._running = True
        self._hub = hub or get_hub()
        self._async = None

        # Startup the async watcher to notify the gevent loop from other
        # threads when using gevent 1.0
        if not _using_libevent:
            self._async = self._hub.loop.async()

        # Spawn our worker greenlets, we have
        # - A completion worker for when values come back to be set on
        #   the AsyncResult object
        # - A callback worker for watch events to be called
        # - A session worker for session events to be called
        self._create_greenlet_worker(self.completion_queue)
        self._create_greenlet_worker(self.callback_queue)
        self._create_greenlet_worker(self.session_queue)

    def _create_greenlet_worker(self, queue):
        def greenlet_worker():
            while self._running:
                # We timeout after 1 and repeat so that we can gracefully
                # shutdown if self_running is set to false
                try:
                    func = queue.get(timeout=1)
                    func()
                except Empty:
                    continue
        gevent.spawn(greenlet_worker)

    def wake(self):
        """Wake the gevent hub the appropriate way"""
        if _using_libevent:
            # Wake gevent wait/gets
            os.write(_core_pipe_write, '\0')
        else:
            self._async.send()

    def event_object(self):
        """Create an appropriate Event object"""
        return gevent.event.Event()

    def lock_object(self):
        """Create an appropriate Lock object"""
        return gevent.thread.allocate_lock()

    def async_result(self):
        """Create a :class:`AsyncResult` instance

        The :class:`AsyncResult` instance will have its completion
        callbacks executed in the thread the :class:`SequentialGeventHandler`
        is created in (which should be the gevent/main thread).

        """
        return AsyncResult(self)

    def spawn(self, func, *args, **kwargs):
        """Spawn a function to run asynchronously"""
        gevent.spawn(func, *args, **kwargs)

    def dispatch_callback(self, callback):
        """Dispatch to the callback object

        The callback is put on separate queues to run depending on the type
        as documented for the :class:`SequentialGeventHandler`.

        """
        if callback.type == 'session':
            self.session_queue.put(lambda: callback.func(*callback.args))
        else:
            self.callback_queue.put(lambda: callback.func(*callback.args))
        self.wake()



import uuid
import threading

import zookeeper
from nose.tools import eq_

from kazoo.exceptions import CancelledError
from kazoo.testing import KazooTestCase
from kazoo.testing import ZooError
from kazoo.tests.util import wait


class KazooLockTests(KazooTestCase):
    def setUp(self):
        super(KazooLockTests, self).setUp()
        self.lockpath = "/" + uuid.uuid4().hex

        self.condition = threading.Condition()
        self.released = threading.Event()
        self.active_thread = None
        self.cancelled_threads = []

    def _thread_lock_acquire_til_event(self, name, lock, event):
        try:
            with lock:
                with self.condition:
                    eq_(self.active_thread, None)
                    self.active_thread = name
                    self.condition.notify_all()

                event.wait()

                with self.condition:
                    eq_(self.active_thread, name)
                    self.active_thread = None
                    self.condition.notify_all()
            self.released.set()
        except CancelledError:
            with self.condition:
                self.cancelled_threads.append(name)
                self.condition.notify_all()

    def test_lock_one(self):
        lock_name = uuid.uuid4().hex
        lock = self.client.Lock(self.lockpath, lock_name)
        event = threading.Event()

        thread = threading.Thread(target=self._thread_lock_acquire_til_event,
            args=(lock_name, lock, event))
        thread.start()

        lock2_name = uuid.uuid4().hex
        anotherlock = self.client.Lock(self.lockpath, lock2_name)

        # wait for any contender to show up on the lock
        wait(anotherlock.contenders)
        eq_(anotherlock.contenders(), [lock_name])

        with self.condition:
            while self.active_thread != lock_name:
                self.condition.wait()

        # release the lock
        event.set()

        with self.condition:
            while self.active_thread:
                self.condition.wait()
        self.released.wait()

    def test_lock(self):
        threads = []
        names = ["contender" + str(i) for i in range(5)]

        contender_bits = {}

        for name in names:
            e = threading.Event()

            l = self.client.Lock(self.lockpath, name)
            t = threading.Thread(target=self._thread_lock_acquire_til_event,
                args=(name, l, e))
            contender_bits[name] = (t, e)
            threads.append(t)

        # acquire the lock ourselves first to make the others line up
        lock = self.client.Lock(self.lockpath, "test")
        lock.acquire()

        for t in threads:
            t.start()

        # wait for everyone to line up on the lock
        wait(lambda: len(lock.contenders()) == 6)
        contenders = lock.contenders()

        eq_(contenders[0], "test")
        contenders = contenders[1:]
        remaining = list(contenders)

        # release the lock and contenders should claim it in order
        lock.release()

        for contender in contenders:
            thread, event = contender_bits[contender]

            with self.condition:
                while not self.active_thread:
                    self.condition.wait()
                eq_(self.active_thread, contender)

            eq_(lock.contenders(), remaining)
            remaining = remaining[1:]

            event.set()

            with self.condition:
                while self.active_thread:
                    self.condition.wait()
            thread.join()

    def test_lock_fail_first_call(self):
        self.add_errors(dict(
            acreate=[True,  # This is our lock node create
                     ZooError('completion', zookeeper.CONNECTIONLOSS, True)
                    ]
        ))

        event1 = threading.Event()
        lock1 = self.client.Lock(self.lockpath, "one")
        thread1 = threading.Thread(target=self._thread_lock_acquire_til_event,
            args=("one", lock1, event1))
        thread1.start()

        # wait for this thread to acquire the lock
        with self.condition:
            if not self.active_thread:
                self.condition.wait(5)
                eq_(self.active_thread, "one")
        eq_(lock1.contenders(), ["one"])
        event1.set()
        thread1.join()

    def test_lock_cancel(self):
        event1 = threading.Event()
        lock1 = self.client.Lock(self.lockpath, "one")
        thread1 = threading.Thread(target=self._thread_lock_acquire_til_event,
            args=("one", lock1, event1))
        thread1.start()

        # wait for this thread to acquire the lock
        with self.condition:
            if not self.active_thread:
                self.condition.wait(5)
                eq_(self.active_thread, "one")

        client2 = self._get_client()
        client2.connect()
        event2 = threading.Event()
        lock2 = client2.Lock(self.lockpath, "two")
        thread2 = threading.Thread(target=self._thread_lock_acquire_til_event,
            args=("two", lock2, event2))
        thread2.start()

        # this one should block in acquire. check that it is a contender
        wait(lambda: len(lock2.contenders()) > 1)
        eq_(lock2.contenders(), ["one", "two"])

        lock2.cancel()
        with self.condition:
            if not "two" in self.cancelled_threads:
                self.condition.wait()
                assert "two" in self.cancelled_threads

        eq_(lock2.contenders(), ["one"])

        thread2.join()
        event1.set()
        thread1.join()
        client2.stop()



"""Kazoo Security

This modules includes helper functions to create digest ACL's appropriate for
use with Zookeeper.

"""
import hashlib

import zookeeper


__all__ = ("make_digest_acl", )


def make_digest_acl_credential(username, password):
    """Create a SHA1 digest credential"""
    credential = "%s:%s" % (username, password)
    cred_hash = hashlib.sha1(credential).digest().encode('base64').strip()
    return "%s:%s" % (username, cred_hash)


def make_acl(scheme, credential, read=False, write=False,
             create=False, delete=False, admin=False, all=False):
    """Given a scheme and credential, return an ACL dict appropriate for
    Zookeeper"""
    if all:
        permissions = ACLPermission.ALL
    else:
        permissions = 0
        if read:
            permissions |= ACLPermission.READ
        if write:
            permissions |= ACLPermission.WRITE
        if create:
            permissions |= ACLPermission.CREATE
        if delete:
            permissions |= ACLPermission.DELETE
        if admin:
            permissions |= ACLPermission.ADMIN

    return dict(scheme=scheme, id=credential, perms=permissions)


def make_digest_acl(username, password, read=False, write=False,
                    create=False, delete=False, admin=False, all=False):
    """Create a digest ACL for Zookeeper with the given permissions"""
    cred = make_digest_acl_credential(username, password)
    return make_acl("digest", cred, read=read, write=write, create=create,
        delete=delete, admin=admin, all=all)


class ACLPermission(object):
    """ACL Permission object"""
    READ = zookeeper.PERM_READ
    WRITE = zookeeper.PERM_WRITE
    CREATE = zookeeper.PERM_CREATE
    DELETE = zookeeper.PERM_DELETE
    ADMIN = zookeeper.PERM_ADMIN
    ALL = zookeeper.PERM_ALL



import unittest

from nose.tools import eq_
import zookeeper


class TestACL(unittest.TestCase):
    def _makeOne(self, *args, **kwargs):
        from kazoo.security import make_acl
        return make_acl(*args, **kwargs)

    def test_read_acl(self):
        acl = self._makeOne("digest", ":", read=True)
        eq_(acl['perms'] & zookeeper.PERM_READ, zookeeper.PERM_READ)

    def test_all_perms(self):
        acl = self._makeOne("digest", ":", write=True, create=True,
                            delete=True, admin=True)
        for perm in [zookeeper.PERM_WRITE, zookeeper.PERM_CREATE,
                     zookeeper.PERM_DELETE, zookeeper.PERM_ADMIN]:
            eq_(acl['perms'] & perm, perm)



import uuid
import time


from nose.tools import eq_
from nose.tools import raises

from kazoo.testing import KazooTestCase


class KazooChildrenWatcherTests(KazooTestCase):
    def setUp(self):
        super(KazooChildrenWatcherTests, self).setUp()
        self.path = "/" + uuid.uuid4().hex

    def _makeOne(self, *args, **kwargs):
        from kazoo.recipe.partitioner import ChildrenWatcher
        return ChildrenWatcher(*args, **kwargs)

    def test_watch(self):
        self.client.ensure_path(self.path)
        watcher = self._makeOne(self.client, self.path, 0.1)
        result = watcher.start()
        children, asy = result.get()
        eq_(len(children), 0)
        eq_(asy.ready(), False)

        self.client.create(self.path + '/' + 'fred', '0')
        asy.get(timeout=1)
        eq_(asy.ready(), True)

    def test_exception(self):
        from kazoo.exceptions import NoNodeException
        watcher = self._makeOne(self.client, self.path, 0.1)
        result = watcher.start()

        @raises(NoNodeException)
        def testit():
            result.get()
        testit()

    def test_watch_iterations(self):
        self.client.ensure_path(self.path)
        watcher = self._makeOne(self.client, self.path, 0.2)
        result = watcher.start()
        eq_(result.ready(), False)

        time.sleep(0.1)
        self.client.create(self.path + '/' + uuid.uuid4().hex, '0')
        eq_(result.ready(), False)
        time.sleep(0.1)
        eq_(result.ready(), False)
        self.client.create(self.path + '/' + uuid.uuid4().hex, '0')
        time.sleep(0.1)
        eq_(result.ready(), False)

        children, asy = result.get()
        eq_(len(children), 2)


class KazooPartitionerTests(KazooTestCase):
    def setUp(self):
        super(KazooPartitionerTests, self).setUp()
        self.path = "/" + uuid.uuid4().hex

    def test_party_of_one(self):
        partitioner = self.client.SetPartitioner(
            self.path, set=(1, 2, 3), time_boundary=0.2)
        eq_(partitioner.acquired, False)
        partitioner.wait_for_acquire(1)
        eq_(partitioner.acquired, True)
        eq_(list(partitioner), [1, 2, 3])
        partitioner.finish()

    def test_party_of_two(self):
        partitioners = [self.client.SetPartitioner(self.path, (1, 2),
                        identifier="p%s" % i, time_boundary=0.2)
                        for i in range(2)]

        eq_(partitioners[0].acquired, False)
        partitioners[0].wait_for_acquire(1)
        partitioners[1].wait_for_acquire(1)
        eq_(list(partitioners[0]), [1])
        eq_(list(partitioners[1]), [2])
        partitioners[0].finish()
        time.sleep(0.1)
        eq_(partitioners[1].release, True)
        partitioners[1].finish()

    def test_party_expansion(self):
        partitioners = [self.client.SetPartitioner(self.path, (1, 2, 3),
                        identifier="p%s" % i, time_boundary=0.2)
                        for i in range(2)]

        eq_(partitioners[0].acquired, False)
        partitioners[0].wait_for_acquire(1)
        partitioners[1].wait_for_acquire(1)

        eq_(list(partitioners[0]), [1, 3])
        eq_(list(partitioners[1]), [2])

        # Add another partition, wait till they settle
        partitioners.append(self.client.SetPartitioner(self.path, (1, 2, 3),
                            identifier="p2", time_boundary=0.2))
        time.sleep(0.1)
        eq_(partitioners[0].release, True)
        for p in partitioners[:-1]:
            p.release_set()

        for p in partitioners:
            p.wait_for_acquire(1)

        eq_(list(partitioners[0]), [1])
        eq_(list(partitioners[1]), [2])
        eq_(list(partitioners[2]), [3])

        for p in partitioners:
            p.finish()

    def test_more_members_than_set_items(self):
        partitioners = [self.client.SetPartitioner(self.path, (1,),
                        identifier="p%s" % i, time_boundary=0.2)
                        for i in range(2)]

        eq_(partitioners[0].acquired, False)
        partitioners[0].wait_for_acquire(1)
        partitioners[1].wait_for_acquire(1)

        eq_(list(partitioners[0]), [1])
        eq_(list(partitioners[1]), [])

        for p in partitioners:
            p.finish()



import unittest

from nose.tools import eq_
import zookeeper


class TestExceptions(unittest.TestCase):
    def _makeOne(self, *args, **kwargs):
        from kazoo.exceptions import err_to_exception
        return err_to_exception(*args, **kwargs)

    def test_error_translate(self):
        exc = self._makeOne(zookeeper.SYSTEMERROR)
        assert isinstance(exc, zookeeper.SystemErrorException)

    def test_error_with_message(self):
        exc = self._makeOne(zookeeper.NONODE, msg="oops")
        assert isinstance(exc, zookeeper.NoNodeException)
        eq_(str(exc), "no node: oops")

    def test_generic_error_code(self):
        exc = self._makeOne(-200)
        assert isinstance(exc, Exception)

    def test_zookeeper_ok(self):
        exc = self._makeOne(zookeeper.OK)
        eq_(exc, None)

    def test_not_error_code(self):
        exc = self._makeOne("this needs to be an int")
        assert isinstance(exc, Exception)



"""Kazoo testing harnesses"""
import atexit
import logging
import os
import uuid
from collections import namedtuple
import unittest

import zookeeper
from kazoo.client import Callback
from kazoo.client import KazooClient
from kazoo.testing.common import ZookeeperCluster

log = logging.getLogger(__name__)

CLUSTER = None


def get_global_cluster():
    global CLUSTER
    if CLUSTER is None:
        ZK_HOME = os.environ.get("ZOOKEEPER_PATH")
        assert ZK_HOME, (
            "ZOOKEEPER_PATH environment variable must be defined.\n "
            "For deb package installations this is /usr/share/java")

        CLUSTER = ZookeeperCluster(ZK_HOME)
        atexit.register(lambda cluster: cluster.terminate(), CLUSTER)
    return CLUSTER


class ZooError(namedtuple('ZooError', ('when', 'exception', 'allow'))):
    """A Zookeeper Error to throw instead of or in addition to executing the
    zookeeper command

    Since the :class:`KazooClient` implements most of the zookeeper commands
    using the async calls, the exception could occur during the call itself
    (which is rare), or on the completion.

    .. attribute:: when

        When the exception should be tossed, can be set to 'call' or
        'completion'.

    .. attribute:: exception

        The exception instance or zookeeper exception code (an int).

    .. attribute:: allow

        Boolean indicating if the zookeeper function should be called
        even though an exception is being returned. This is useful for
        reproducing rare events in Zookeeper where a connection is lost
        before a response is returned but the command ran successfully.

    """


class ZookeeperErrors(object):
    """A Zookeeper proxy with errors"""
    def __init__(self, errors, handler):
        """Create a :class:`ZookeeperErrors` object that throws the
        desired errors upon calls

        :param errors: A dict structure keyed by the zookeeper function
                       to intercept, with the value being a list of
                       :class:`ZooError`s to throw. When the list is empty
                       calls will proceed as normal, include ``True`` in a
                       place to allow a call to pass if desired as well.
        :param handler: The handler instance used by the KazooClient.

        Example::

            errors = dict(
                acreate=[
                    ZooError('completion', zookeeper.CONNECTIONLOSS, False),
                    True,
                    ZooError('call', SystemError(), False)
                ]
            )

        .. warning::

            Care should be taken when simulating errors that your test code
            executes the zookeeper commands in a reliably deterministic manner.

        """
        for k, v in errors.items():
            v.reverse()
        self.errors = errors
        self.handler = handler

    def _intercept_completion(self):
        # Completion callback is always at the end
        asy = self.handler.async_result()

        def new_completion(handle, code, *args):
            log.info("Intercept completion, handle: %s, "
                     "code: %s, args: %s", handle, code, args)
            asy.set()
        return asy, new_completion

    def _intercept_watcher(self):
        def new_watcher(*args):
            log.info("Intercept watcher, args: %s", args)
        return new_watcher

    def _call_exception(self, err, name, func, *args, **kwargs):
        if err.allow:
            # Strip completion/watch callbacks, replace with dummies
            asy = False
            if callable(args[-1]):
                asy, args[-1] = self._intercept_completion()
            if callable(args[-2]):
                args[-2] = self._intercept_watcher()
            log.debug("Calling actual function: %s", name)
            func(*args, **kwargs)
            if asy:
                asy.wait()
            log.debug("Raising desired exception: %s", err.exception)
            raise err.exception
        else:
            log.debug("Raising desired exception on call: %s",
                err.exception)
            raise err.exception

    def _completion_exception(self, err, name, func, *args, **kwargs):
        # Ensure args is a list
        if isinstance(args, tuple):
            args = list(args)

        if callable(args[-1]):
            # Replace completion callback with one that returns the
            # exception
            old_complete = args[-1]
            args[-1] = lambda handle, code, *new_args: old_complete(
                handle, err.exception, *new_args)
        if callable(args[-2]):
            # Strip the watch callback
            args[-2] = lambda *args: 1
            # Call it if we're supposed to
        if err.allow:
            log.debug("Calling actual function: %s", name)
            func(*args, **kwargs)
        else:
            # Make sure the completion callback is called
            log.debug("Returning desired exception on call: %s",
                err.exception)
            self.handler.dispatch_callback(
                Callback('completion', func, (None, err.exception))
            )
        return zookeeper.OK

    def __getattr__(self, name):
        func = getattr(zookeeper, name)
        err = self.errors.get(name)
        if err:
            err = err.pop()
        if not isinstance(err, ZooError):
            return func

        def func_wrapper(*args, **kwargs):
            if err.when == 'call':
                return self._call_exception(err, name, func, *args, **kwargs)
            else:
                return self._completion_exception(err, name, func, *args,
                    **kwargs)
        return func_wrapper


class KazooTestHarness(object):
    """Harness for testing code that uses Kazoo

    This object can be used directly or as a mixin. It supports starting
    and stopping a complete ZooKeeper cluster locally and provides an
    API for simulating errors and expiring sessions.

    Example::

        class MyTestCase(unittest.TestCase, KazooTestHarness):
            def setUp(self):
                self.setup_zookeeper()

                # additional test setup

            def tearDown(self):
                self.teardown_zookeeper()

            def test_something(self):
                something_that_needs_a_kazoo_client(self.client)

            def test_something_else(self):
                something_that_needs_zk_servers(self.servers)

    """

    def __init__(self):
        self.client = None

    @property
    def cluster(self):
        return get_global_cluster()

    @property
    def servers(self):
        return ",".join([s.address for s in self.cluster])

    def _get_nonchroot_client(self):
        return KazooClient(self.servers)

    def _get_client(self, **kwargs):
        return KazooClient(self.hosts, **kwargs)

    def expire_session(self, client_id=None):
        """Force ZK to expire a client session

        :param client_id: id of client to expire. If unspecified, the id of
                          self.client will be used.

        """
        client_id = client_id or self.client.client_id
        client = KazooClient(self.cluster[1].address, client_id=client_id)
        client.connect()
        client.stop()

    def setup_zookeeper(self):
        """Create a ZK cluster and chrooted :class:`KazooClient`

        The cluster will only be created on the first invocation and won't be
        fully torn down until exit.
        """
        zookeeper.deterministic_conn_order(True)
        if not self.cluster[0].running:
            self.cluster.start()
        namespace = "/kazootests" + uuid.uuid4().hex
        self.hosts = self.servers + namespace

        self.client = self._get_client()
        self.client.connect()
        self.client.ensure_path("/")

    def teardown_zookeeper(self):
        """Clean up any ZNodes created during the test
        """
        if not self.cluster[0].running:
            self.cluster.start()

        if self.client.connected:
            self.client.stop()

        client = self._get_client()
        client.connect()
        client.delete('/', recursive=True)
        client.stop()

    def add_errors(self, errors):
        self.client.zookeeper = ZookeeperErrors(errors, self.client._handler)


class KazooTestCase(unittest.TestCase, KazooTestHarness):
    def setUp(self):
        self.setup_zookeeper()

    def tearDown(self):
        self.teardown_zookeeper()



import threading
import uuid

from nose.tools import eq_

from kazoo.testing import KazooTestCase


class KazooDataWatcherTests(KazooTestCase):
    def setUp(self):
        super(KazooDataWatcherTests, self).setUp()
        self.path = "/" + uuid.uuid4().hex

    def test_child_watcher(self):
        self.client.ensure_path(self.path)
        update = threading.Event()
        data = [True]

        @self.client.DataWatch(self.path)
        def changed(d, stat):
            data.pop()
            data.append(d)
            update.set()

        update.wait()
        eq_(data, [""])
        update.clear()

        self.client.set(self.path, 'fred')
        update.wait()
        eq_(data[0], 'fred')
        update.clear()

    def test_func_stops(self):
        self.client.ensure_path(self.path)
        update = threading.Event()
        data = [True]

        fail_through = []

        @self.client.DataWatch(self.path)
        def changed(d, stat):
            data.pop()
            data.append(d)
            update.set()
            if fail_through:
                return False

        update.wait()
        eq_(data, [""])
        update.clear()

        fail_through.append(True)
        self.client.set(self.path, 'fred')
        update.wait()
        eq_(data[0], 'fred')
        update.clear()

        self.client.set(self.path, 'asdfasdf')
        update.wait(0.5)
        eq_(data[0], 'fred')

        d, stat = self.client.get(self.path)
        eq_(d, 'asdfasdf')


class KazooChildrenWatcherTests(KazooTestCase):
    def setUp(self):
        super(KazooChildrenWatcherTests, self).setUp()
        self.path = "/" + uuid.uuid4().hex

    def test_child_watcher(self):
        self.client.ensure_path(self.path)
        update = threading.Event()
        all_children = ['fred']

        @self.client.ChildrenWatch(self.path)
        def changed(children):
            while all_children:
                all_children.pop()
            all_children.extend(children)
            update.set()

        update.wait()
        eq_(all_children, [])
        update.clear()

        self.client.create(self.path + '/' + 'smith', '0')
        update.wait()
        eq_(all_children, ['smith'])
        update.clear()

        self.client.create(self.path + '/' + 'george', '0')
        update.wait()
        eq_(sorted(all_children), ['george', 'smith'])

    def test_func_stops(self):
        self.client.ensure_path(self.path)
        update = threading.Event()
        all_children = ['fred']

        fail_through = []

        @self.client.ChildrenWatch(self.path)
        def changed(children):
            while all_children:
                all_children.pop()
            all_children.extend(children)
            update.set()
            if fail_through:
                return False

        update.wait()
        eq_(all_children, [])
        update.clear()

        fail_through.append(True)
        self.client.create(self.path + '/' + 'smith', '0')
        update.wait()
        eq_(all_children, ['smith'])
        update.clear()

        self.client.create(self.path + '/' + 'george', '0')
        update.wait(0.5)
        eq_(all_children, ['smith'])

    def test_child_watch_session_loss(self):
        self.client.ensure_path(self.path)
        update = threading.Event()
        all_children = ['fred']

        @self.client.ChildrenWatch(self.path)
        def changed(children):
            while all_children:
                all_children.pop()
            all_children.extend(children)
            update.set()

        update.wait()
        eq_(all_children, [])
        update.clear()

        self.client.create(self.path + '/' + 'smith', '0')
        update.wait()
        eq_(all_children, ['smith'])
        update.clear()
        self.expire_session()

        eq_(update.is_set(), False)

        self.client.retry(self.client.create,
                          self.path + '/' + 'george', '0')
        update.wait()
        eq_(sorted(all_children), ['george', 'smith'])

    def test_child_stop_on_session_loss(self):
        self.client.ensure_path(self.path)
        update = threading.Event()
        all_children = ['fred']

        @self.client.ChildrenWatch(self.path, allow_session_lost=False)
        def changed(children):
            while all_children:
                all_children.pop()
            all_children.extend(children)
            update.set()

        update.wait()
        eq_(all_children, [])
        update.clear()

        self.client.create(self.path + '/' + 'smith', '0')
        update.wait()
        eq_(all_children, ['smith'])
        update.clear()
        self.expire_session()

        eq_(update.is_set(), False)

        self.client.retry(self.client.create,
                          self.path + '/' + 'george', '0')
        update.wait(0.5)
        eq_(update.is_set(), False)
        eq_(all_children, ['smith'])

        children = self.client.get_children(self.path)
        eq_(sorted(children), ['george', 'smith'])



"""Kazoo Interfaces"""
from zope.interface import (
    Attribute,
    Interface,
)

# public API


class IHandler(Interface):
    """A Callback Handler for Zookeeper completion and watch callbacks

    This object must implement several methods responsible for determining
    how completion / watch callbacks are handled as well as the method for
    calling :class:`IAsyncResult` callback functions.

    The Handler should document how callbacks are called for:

    * Zookeeper completion events
    * Zookeeper session events
    * Zookeeper watch events

    """
    name = Attribute(
        """Human readable name of the Handler interface""")

    timeout_exception = Attribute(
        """Exception class that should be thrown and captured if a result
        is not available within the given time""")

    sleep_func = Attribute(
        """Appropriate sleep function that can be called with a single
        argument and sleep.""")

    def event_object():
        """Return an appropriate object that implements Python's
        threading.Event API"""

    def lock_object():
        """Return an appropriate object that implements Python's
        threading.Lock API"""

    def async_result():
        """Return an instance that conforms to the :class:`~IAsyncResult`
        interface appropriate for this handler"""

    def spawn(func, *args, **kwargs):
        """Spawn a function to run asynchronously

        :param args: args to call the function with
        :param kwargs: keyword args to call the function with

        This method should return immediately and execute the function
        with the provided args and kwargs in an asynchronous manner.

        """

    def dispatch_callback(callback):
        """Dispatch to the callback object

        :param callback: A :class:`~kazoo.client.Callback` object to be
                         called

        """


class IAsyncResult(Interface):
    """An Async Result object that can be queried for a value that has been
    set asyncronously

    This object is modeled on the ``gevent`` AsyncResult object.

    The implementation must account for the fact that the :meth:`set` and
    :meth:`set_exception` methods will be called from within the Zookeeper
    thread which may require extra care under asynchronous environments.

    """
    value = Attribute(
        """Holds the value passed to :meth:`set` if :meth:`set` was called.
        Otherwise `None`""")

    exception = Attribute(
        """Holds the exception instance passed to :meth:`set_exception` if
        :meth:`set_exception` was called. Otherwise `None`""")

    def ready():
        """Return `True` if and only if it holds a value or an exception"""

    def successful():
        """Return `True` if and only if it is ready and holds a value"""

    def set(value=None):
        """Store the value. Wake up the waiters.

        :param value: Value to store as the result

        Any waiters blocking on :meth:`get` or :meth:`wait` are woken up.
        Sequential calls to :meth:`wait` and :meth:`get` will not block at
        all."""

    def set_exception(exception):
        """Store the exception. Wake up the waiters.

        :param exception: Exception to raise when fetching the value

        Any waiters blocking on :meth:`get` or :meth:`wait` are woken up.
        Sequential calls to :meth:`wait` and :meth:`get` will not block at
        all."""

    def get(block=True, timeout=None):
        """Return the stored value or raise the exception

        :param block: Whether this method should block or return immediately
        :type block: bool
        :param timeout: How long to wait for a value when `block` is `True`
        :type timeout: float

        If this instance already holds a value / an exception, return / raise
        it immediately. Otherwise, block until :meth:`set` or
        :meth:`set_exception` has been called or until the optional timeout
        occurs."""

    def get_nowait():
        """Return the value or raise the exception without blocking.

        If nothing is available, raise the Timeout exception class on the
        associated :class:`IHandler` interface."""

    def wait(timeout=None):
        """Block until the instance is ready.

        :param timeout: How long to wait for a value when `block` is `True`
        :type timeout: float

        If this instance already holds a value / an exception, return / raise
        it immediately. Otherwise, block until :meth:`set` or
        :meth:`set_exception` has been called or until the optional timeout
        occurs."""

    def rawlink(callback):
        """Register a callback to call when a value or an exception is set

        :param callback: A callback function to call after :meth:`set` or
                         :meth:`set_exception` has been called. This function
                         will be passed a single argument, this instance.
        :type callback: func

        """

    def unlink(callback):
        """Remove the callback set by :meth:`rawlink`

        :param callback: A callback function to remove
        :type callback: func

        """



"""Zookeeper Locking Implementations

Error Handling
==============

It's highly recommended to add a state listener with
:meth:`~KazooClient.add_listener` and watch for
:attr:`~KazooState.LOST` and :attr:`~KazooState.SUSPENDED` state
changes and re-act appropriately. In the event that a
:attr:`~KazooState.LOST` state occurs, its certain that the lock
has been lost.

"""
import uuid

from kazoo.retry import ForceRetryError
from kazoo.exceptions import CancelledError
from kazoo.exceptions import NoNodeException


class Lock(object):
    """Kazoo Basic Lock

    Example usage with a :class:`~kazoo.client.KazooClient` instance::

        zk = KazooClient()
        lock = zk.Lock("/lockpath", "my-identifier")
        with lock:  # blocks waiting for lock acquisition
            # do something with the lock

    """
    _NODE_NAME = '__lock__'

    def __init__(self, client, path, identifier=None):
        """Create a Kazoo Lock

        :param client: A :class:`~kazoo.client.KazooClient` instance
        :param path: The lock path to use
        :param identifier: Name to use for this lock contender. This
                           can be useful for querying to see who the
                           current lock contenders are.

        """
        self.client = client
        self.path = path

        # some data is written to the node. this can be queried via
        # contenders() to see who is contending for the lock
        self.data = str(identifier or "")

        self.wake_event = client._handler.event_object()

        # props to Netflix Curator for this trick. It is possible for our
        # create request to succeed on the server, but for a failure to
        # prevent us from getting back the full path name. We prefix our
        # lock name with a uuid and can check for its presence on retry.
        self.prefix = uuid.uuid4().hex + self._NODE_NAME
        self.create_path = self.path + "/" + self.prefix

        self.create_tried = False
        self.is_acquired = False
        self.assured_path = False
        self.cancelled = False

    def cancel(self):
        """Cancel a pending lock acquire"""
        self.cancelled = True
        self.wake_event.set()

    def acquire(self):
        """Acquire the mutex, blocking until it is obtained"""
        try:
            self.client.retry(self._inner_acquire)
            self.is_acquired = True
        except Exception:
            # if we did ultimately fail, attempt to clean up
            self._best_effort_cleanup()
            self.cancelled = False
            raise

    def _inner_acquire(self):
        self.wake_event.clear()

        # make sure our election parent node exists
        if not self.assured_path:
            self.client.ensure_path(self.path)

        node = None
        if self.create_tried:
            node = self._find_node()
        else:
            self.create_tried = True

        if not node:
            node = self.client.create(self.create_path, self.data,
                ephemeral=True, sequence=True)
            # strip off path to node
            node = node[len(self.path) + 1:]

        self.node = node

        while True:

            # bail out with an exception if cancellation has been requested
            if self.cancelled:
                raise CancelledError()

            children = self._get_sorted_children()

            try:
                our_index = children.index(node)
            except ValueError:
                # somehow we aren't in the children -- probably we are
                # recovering from a session failure and our ephemeral
                # node was removed
                raise ForceRetryError()

            #noinspection PySimplifyBooleanCheck
            if our_index == 0:
                # we have the lock
                return True

            # otherwise we are in the mix. watch predecessor and bide our time
            predecessor = self.path + "/" + children[our_index - 1]
            if self.client.exists(predecessor, self._watch_predecessor):
                self.wake_event.wait()

    def _watch_predecessor(self, event):
        self.wake_event.set()

    def _get_sorted_children(self):
        children = self.client.get_children(self.path)

        # can't just sort directly: the node names are prefixed by uuids
        lockname = self._NODE_NAME
        children.sort(key=lambda c: c[c.find(lockname) + len(lockname):])
        return children

    def _find_node(self):
        children = self.client.get_children(self.path)
        for child in children:
            if child.startswith(self.prefix):
                return child
        return None

    def _best_effort_cleanup(self):
        try:
            node = self._find_node()
            if node:
                self.client.delete(self.path + "/" + node)
        except Exception:
            pass

    def release(self):
        """Release the mutex immediately"""
        return self.client.retry(self._inner_release)

    def _inner_release(self):
        if not self.is_acquired:
            return False

        self.client.delete(self.path + "/" + self.node)

        self.is_acquired = False
        self.node = None

        return True

    def contenders(self):
        """Return an ordered list of the current contenders for the lock

        .. note::

            If the contenders did not set an identifier, it will appear
            as a blank string.

        """
        # make sure our election parent node exists
        if not self.assured_path:
            self.client.ensure_path(self.path)

        children = self._get_sorted_children()

        contenders = []
        for child in children:
            try:
                data, stat = self.client.get(self.path + "/" + child)
                contenders.append(data)
            except NoNodeException:
                pass
        return contenders

    def __enter__(self):
        self.acquire()

    def __exit__(self, exc_type, exc_value, traceback):
        self.release()



__version__ = '0.2b1'

import os

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.rst')) as f:
    README = f.read()
with open(os.path.join(here, 'CHANGES.rst')) as f:
    CHANGES = f.read()

install_requires = [
    'zope.interface >= 3.8.0',  # has zope.interface.registry
]

tests_require = install_requires + [
    'coverage',
    'docutils',
    'gevent',
    'mock',
    'nose',
    'repoze.sphinx.autointerface',
    'Sphinx',
    ]

on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
if on_rtd:
    install_requires.extend([
        'gevent',
        'repoze.sphinx.autointerface',
    ])

setup(
    name='kazoo',
    version=__version__,
    description='Higher Level Zookeeper Client',
    long_description=README + '\n\n' + CHANGES,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Topic :: Communications",
        "Topic :: System :: Distributed Computing",
        "Topic :: System :: Networking",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords='zookeeper lock leader configuration',
    author="Nimbus team, Zope Corporation, Mozilla Foundation",
    author_email="",
    url="http://kazoo.readthedocs.org/",
    license="Apache 2.0",
    packages=find_packages(),
    test_suite="kazoo.tests",
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require={
        'test': tests_require,
    },
    entry_points="""
    """
)



import uuid

from nose.tools import eq_

from kazoo.testing import KazooTestCase


class KazooPartyTests(KazooTestCase):
    def setUp(self):
        super(KazooPartyTests, self).setUp()
        self.path = "/" + uuid.uuid4().hex

    def test_party(self):
        parties = [self.client.Party(self.path, "p%s" % i)
                    for i in range(5)]

        one_party = parties[0]

        eq_(list(one_party), [])
        eq_(len(one_party), 0)

        participants = set()
        for party in parties:
            party.join()
            participants.add(party.data)

            eq_(set(party), participants)
            eq_(len(party), len(participants))

        for party in parties:
            party.leave()
            participants.remove(party.data)

            eq_(set(party), participants)
            eq_(len(party), len(participants))


class KazooShallowPartyTests(KazooTestCase):
    def setUp(self):
        super(KazooShallowPartyTests, self).setUp()
        self.path = "/" + uuid.uuid4().hex

    def test_party(self):
        parties = [self.client.ShallowParty(self.path, "p%s" % i)
                    for i in range(5)]

        one_party = parties[0]

        eq_(list(one_party), [])
        eq_(len(one_party), 0)

        participants = set()
        for party in parties:
            party.join()
            participants.add(party.data)

            eq_(set(party), participants)
            eq_(len(party), len(participants))

        for party in parties:
            party.leave()
            participants.remove(party.data)

            eq_(set(party), participants)
            eq_(len(party), len(participants))



"""Child and Data watching higher level API's

"""
import logging

from kazoo.client import KazooState

log = logging.getLogger(__name__)


class DataWatch(object):
    """Watches a node for data updates and calls the specified
    function each time it changes

    The function will also be called the very first time its
    registered to get the data.

    Returning `False` from the registered function will disable
    future data change calls.

    Example with client:

    .. code-block:: python

        @client.DataWatch('/path/to/watch')
        def my_func(data, stat):
            print "Data is %s" % data
            print "Version is %s" % stat.version

        # Above function is called immediately and prints

    """
    def __init__(self, client, path, func=None,
                 allow_session_lost=True):
        """Create a children watcher for a path

        :param client: A zookeeper client
        :type client: :class:`~kazoo.client.KazooClient`
        :param path: The path to watch for children on
        :type path: str
        :param func: Function to call initially and every time the
                     children change. `func` will be called with a
                     tuple, the value of the node and a
                     :class:`~kazoo.client.ZnodeStat` instance
        :type func: callable
        :param allow_session_lost: Whether the watch should be
                                   re-registered if the zookeeper
                                   session is lost.
        :type allow_session_lost: bool

        The path must already exist for the children watcher to
        run.

        """
        self._client = client
        self._path = path
        self._func = func
        self._stopped = False
        self._watch_established = False
        self._allow_session_lost = allow_session_lost
        self._run_lock = client._handler.lock_object()
        self._prior_data = ()

        # Register our session listener if we're going to resume
        # across session losses
        if func:
            if allow_session_lost:
                self._client.add_listener(self._session_watcher)
            self._get_data()

    def __call__(self, func):
        """Callable version for use as a decorator

        :param func: Function to call initially and every time the
                     children change. `func` will be called with a
                     tuple, the value of the node and a
                     :class:`~kazoo.client.ZnodeStat` instance
        :type func: callable

        """
        self._func = func

        if self._allow_session_lost:
            self._client.add_listener(self._session_watcher)
        self._get_data()
        return func

    def _get_data(self):
        with self._run_lock:  # Ensure this runs one at a time
            if self._stopped:
                return

            data, stat = self._client.retry(self._client.get,
                                            self._path, self._watcher)
            if not self._watch_established:
                self._watch_established = True

                # If we already had data, and it hasn't changed, this is a
                # session re-establishment and nothing changed, don't call the
                # func
                if self._prior_data and \
                   self._prior_data[1].mzxid == stat.mzxid:
                    return

            self._prior_data = data, stat

            try:
                if self._func(data, stat) is False:
                    self._stopped = True
            except Exception as exc:
                log.exception(exc)
                raise

    def _watcher(self, event):
        self._get_data()

    def _session_watcher(self, state):
        if state == KazooState.LOST:
            self._watch_established = False
        elif state == KazooState.CONNECTED and \
             not self._watch_established and not self._stopped:
            self._get_data()


class ChildrenWatch(object):
    """Watches a node for children updates and calls the specified
    function each time it changes

    The function will also be called the very first time its
    registered to get children.

    Returning `False` from the registered function will disable
    future children change calls.

    Example with client:

    .. code-block:: python

        @client.ChildrenWatch('/path/to/watch')
        def my_func(children):
            print "Children are %s" % children

        # Above function is called immediately and prints children

    """
    def __init__(self, client, path, func=None,
                 allow_session_lost=True):
        """Create a children watcher for a path

        :param client: A zookeeper client
        :type client: :class:`~kazoo.client.KazooClient`
        :param path: The path to watch for children on
        :type path: str
        :param func: Function to call initially and every time the
                     children change. `func` will be called with a
                     single argument, the list of children.
        :type func: callable
        :param allow_session_lost: Whether the watch should be
                                   re-registered if the zookeeper
                                   session is lost.
        :type allow_session_lost: bool

        The path must already exist for the children watcher to
        run.

        """
        self._client = client
        self._path = path
        self._func = func
        self._stopped = False
        self._watch_established = False
        self._allow_session_lost = allow_session_lost
        self._run_lock = client._handler.lock_object()
        self._prior_children = None

        # Register our session listener if we're going to resume
        # across session losses
        if func:
            if allow_session_lost:
                self._client.add_listener(self._session_watcher)
            self._get_children()

    def __call__(self, func):
        """Callable version for use as a decorator

        :param func: Function to call initially and every time the
                     children change. `func` will be called with a
                     single argument, the list of children.
        :type func: callable

        """
        self._func = func

        if self._allow_session_lost:
            self._client.add_listener(self._session_watcher)
        self._get_children()
        return func

    def _get_children(self):
        with self._run_lock:  # Ensure this runs one at a time
            if self._stopped:
                return

            children = self._client.retry(self._client.get_children,
                                          self._path, self._watcher)
            if not self._watch_established:
                self._watch_established = True

                if self._prior_children is not None and \
                   self._prior_children == children:
                    return

            self._prior_children = children

            try:
                if self._func(children) is False:
                    self._stopped = True
            except Exception as exc:
                log.exception(exc)
                raise

    def _watcher(self, event):
        self._get_children()

    def _session_watcher(self, state):
        if state == KazooState.LOST:
            self._watch_established = False
        elif state == KazooState.CONNECTED and \
             not self._watch_established and not self._stopped:
            self._get_children()
